###############################################################################
#POD Part
###############################################################################

=head1 NAME

package I<Alignment> - library used to realize the alignment of sequences
using either a Needleman and Wunsch algorithm, or a global and multiple
alignment using the INFERNAL program developed by Narowcki et al; (2013).
This library is also used to treat

=head1 LIBRARY ORGANIZATION AND FUNCTIONS

	use Alignment;
	LaunchInfernalAlign();
	ExtractDataFromAlignment();
	DefineGaps();
	NeedlemanWunschAlignment();

=head1 DESCRIPTION

This module was defined to manage both types of alignments based on user choices
and also to treat aligned reads defining the number of gaps or extracting data
from a global and multiple alignment.
The documentation was realized with B<Pod>.

=head2 FUNCTIONS

=over 4

=item B<LaunchInfernalAlign()>

This function was developed to launch using the good parameters the INFERNAL
alignment on the temporary file containing both the database and the file that
need treatment. It takes as input the temporary file that will be aligned,
($_[0]), the database folder given as input by the user ($_[1]), and the defined
number of cpu (default or user choice) ($_[2]). As output, it returns the name
of the alignment file.

=item B<ExtractDataFromAlignment()>

This function was developed to read and extract needed data from the global
alignment obtained with INFERNAL. Three hashes references for the data from the
DB or the treated sample were defined and filled and finally returned to the
main program. As input, it takes only the name of the Aligned file stored in the
Temporary folder.

=item B<DefineGaps()>

This function was developed to define easily the number of gaps in each aligned
read from the database and store the info in a hash to avoid multiple treatments
as we have a global and multiple alignment. It takes as input only the reference
of the hash table containing the alignment of the sequences of the DB, and
returns another hash table containing the number of gaps for each sequence.

=item B<NeedlemanWunschAlignment()>

This function was developed to apply the Needleman-Wunsch algorithm of global
alignment between two reads based on user choices. This function will then
return the aligned sequences. It takes as input two sequences, and returns the
aligned sequences between them/

=back

=head1 AUTHORS

Sebastien TERRAT <sebastien.terrat@inra.fr> and Corentin Journay

=head1 TEAM AND NOTES

UMR 1347 INRA/uB/AgroSup - BioCom Team
RELEASED:  30/10/2018

=cut

###############################################################################
#FUNCTIONS AND PROGRAMS
###############################################################################

# Name of the package defined for the main program
package PerlLib::Alignment;
use strict;
use warnings;

###############################################################################
# FUNCTION:		LaunchInfernalAlign
# USAGE:			LaunchInfernalAlign(arguments)
# ARGUMENTS:	- The temporary file that will be aligned ($_[0]).
#							- The database folder given as input by the user ($_[1])
#							- The defined number of cpu (default or user choice) ($_[2])
#							- The defined file name for the alignment ($_[3])
# DETAILS:		This function was developed to launch using the good
#			parameters the INFERNAL alignment on the temporary file
#			containing both the database and the file that need treatment.
#RELEASED: 10/10/2018
###############################################################################
sub LaunchInfernalAlign
{
	#Defined variables
	my $tmp_file = $_[0];
	my $DB_folder = $_[1];
	my $cpu = $_[2];
	my $out_file = $_[3];
	my $filename = "";
	my $dbreads = "";
	my $model = "";
	my $out_data = "";

	#Modification of the filename $out_file to manage the specific case of the alignement of
	#the OTU file from the created database
	if($out_file =~ m/dbreads/)  { $out_file =~ s/dbreads/dbalign/; }
	elsif($out_file =~ m/TMP_seeds/) { $out_file = "Temporary_files/Align_TMP_seeds"; }
	else { $out_file = "Temporary_files/Align_tmp_file"; }

	#Needed command line with parameters to launch alignment steps (--hbanded: needed parameter
	#based on developers information, --dna: to have the alignment is DNA format, not RNA,
	#--matchonly: only include match columns in the output alignment, -ooutformat: to precise
	#the format (Aligned Fasta or AFA) of the output file, --cpu: the number of threads to realize
	#the alignment -o : to precise the filename where the alignment will be stored.
	my $command_align_part = "cmalign --hbanded --outformat pfam --dna --cpu $cpu -o $out_file";

	#We check the chosen type of organism to chose the viable model (bacteia, fungi, algae or archaea).
	if($DB_folder =~ m/BACTERIA/) { $model = "model/bacteria_model_1.1.1.cm"; }
	elsif($DB_folder =~ m/FUNGI/) { $model = "model/fungi_model_1.1.1.cm"; }
	elsif($DB_folder =~ m/ALGAE/) { $model = "model/algae_model_1.1.1.cm"; }
	elsif($DB_folder =~ m/ARCHAEA/) { $model = "model/archaea_model_1.1.1.cm"; }
	else { die "The name of the database folder did not contain the chosen model, please check this problem\n"; }

	#Then, we launch the alignment and return the name of the created file
	$out_data = `$command_align_part $model $tmp_file`;
	return($out_file);
}
#End of function
###############################################################################

###############################################################################
# FUNCTION:		MergeAlignments
# USAGE:			MergeAlignments(arguments)
# ARGUMENTS:	- The temporary file that will be aligned ($_[0]).
#							- The database folder given as input by the user ($_[1])
#							- The name of the output file chosen by the main program ($_[2])
# DETAILS:		This function was developed to merge both alignments from the
#							database and the input reads using the esl-alimerge program
#RELEASED: 10/10/2018
###############################################################################
sub MergeAlignments
{
	#Defined variables
	my $DB_folder = $_[0];
	my $input_file = $_[1];
	my $out_file = $_[2];
	my $dbalign = "";
	my $filename = "";
	my $out_data = "";
	#Needed command line with parameters to launch merging step (--dna: to have the
	#alignment is DNA format, not RNA,-ooutformat: to precise the format (Aligned
	#Fasta or AFA, or PFAM) of the output file, -o : to precise the filename where
	#the alignment will be stored.
	my $command_line_part = "esl-alimerge --dna --outformat pfam -o $out_file";
	#First, we catch and store the name of the dbalign file.
	opendir (DIR, "$DB_folder") || die "can't open the directory $DB_folder: $!";
	LOOP01:while (defined($filename = readdir(DIR))) { if ($filename =~ m/^dbalign/) { $dbalign = $filename; last LOOP01; } }
	closedir(DIR);

	#Then, we launch the merging of the alignments and return the name of the created file
	$out_data = `$command_line_part $DB_folder$dbalign $input_file`;
	return($out_file);
}

###############################################################################
# FUNCTION:		ExtractDataFromAlignment
# USAGE:			ExtractDataFromAlignment(ARGUMENTS)
# ARGUMENTS:	- The temporary file that was previously aligned ($_[0]).
# DETAILS:		This function was developed to read and extract needed data
#			from the global alignment. Three hashes for the data from the
#			DB or the treated sample were defined and filled and finally
#			returned to the main program.
#RELEASED: 10/10/2018
###############################################################################
sub ExtractDataFromAlignment
{
	#Defined variables
	my $align_file = $_[0];
	my $ID = "";
	my $seq = "";
	my @tmp = ();
	my %otu = ();
	my %sample = ();
	my %abundance = ();

	#We open the aligned file that will be treated
	open(ALIGN,"$align_file") || die "Can't open the $align_file, please check this problem.\n";
	while(<ALIGN>)
	{
		#We search for needed data, separating the DB from the sample
		#and storing all data in three hashs tables (%otus, %sample and
		#%abundance).
		if(($_ =~ m/^#/)||($_ eq "\n")||($_ eq "//")) { next; }
		else
		{
			$_ =~ m/([A-Za-z0-9\_\-\=]*)\s*(.*)\n/;
			$ID = $1;
			$seq = $2;;
			if($ID =~ s/^SQ_//)
			{
				$sample{$ID} = $seq;
				@tmp = split("_", $ID);
				$abundance{$ID} = $tmp[1];
			}
			else { $otu{$ID} = $seq; }
		}
	}
	close(ALIGN);
	#We delete the not needed alignment file
	unlink($align_file);
	#We finally return the three filled hashes
	return(\%otu,\%sample,\%abundance);
}
#End of function
###############################################################################

###############################################################################
# FUNCTION:		DefineGaps
# USAGE:			DefineGaps(ARGUMENTS)
# ARGUMENTS:	- The hash table %otu containg the alignment of the database ($_[0]).
# DETAILS:		This function was developed to define easily the number of
#			gaps in each aligned read from the database and store the
#			info in a hash to avoid multiple treatments as we have a
#			global alignment
# RELEASED: 10/10/2018
###############################################################################
sub DefineGaps
{
	#Defined variables
	my $ref_otu = $_[0];
	my $ID = "";
	my $seq = "";
	my $count = 0;
	my %gaps = ();
	#For each read and from the %otu hash table, we count the number of gaps
	while (($ID, $seq) = each(%{$ref_otu}))
	{
		$count = $seq =~ tr/\-\./\-\./;
		$gaps{$ID} = $count;
	}
	#We finally return the %gaps hash table
	return (\%gaps);
}
#End of function
###############################################################################

###############################################################################
# FUNCTION:		NeedlemanWunschAlignment
# USAGE:			NeedlemanWunschAlignment(ARGUMENTS)
# ARGUMENTS:	- The first read that need alignment ($_[0]).
#							- The second read that need alignment ($_[1]).
# DETAILS:		This function was developed to apply the Needleman-Wunsch
#			algorithm of global alignment between two reads based on
#			a user choice. This function will then return the aligned
#			sequences.
#RELEASED: 10/10/2018
###############################################################################
sub NeedlemanWunschAlignment
{
	#Defined variables
	my $seq1 = uc($_[0]);
	my $seq2 = uc($_[1]);
	my %score_matrix = ("A" => {"A"=> 5, "C"=> -4, "G"=> -4, "T"=> -4},
				"C" => {"A"=> -4, "C"=> 5, "G"=> -4, "T"=> -4},
				"G" => {"A"=> -4, "C"=> -4, "G"=> 5, "T"=> -4},
				"T" => {"A"=> -4, "C"=> -4, "G"=> -4, "T"=> 5});
	#Scoring values needed (+1 for a match, and -10 for a gap)
	my $match = 1;
	my $gap = -10;
	#Other variables
	my ($i, $j) = (0,0);
	my ($diagonal_score, $left_score, $up_score) = (0,0,0);
	my ($letter1, $letter2) = ("","");
	my $align1 = "";
	my $align2 = "";
	my @matrix = ();

	#Treatment of both input sequences to delete the non-wanted characters
	$seq1 =~ s/\-//g;
	$seq2 =~ s/\-//g;

	#Initialization of the matrix, and also the first column and the first line
	$matrix[0][0]{score} = 0;
	$matrix[0][0]{pointer} = "none";
	for($j = 1; $j <= length($seq1); $j++) { $matrix[0][$j]{score} = $gap*$j; $matrix[0][$j]{pointer} = "l"; }
	for($i = 1; $i <= length($seq2); $i++) { $matrix[$i][0]{score} = $gap*$i; $matrix[$i][0]{pointer} = "u"; }

	#Using this initialization, the program fill the complete matrices
	for($i = 1; $i <= length($seq2); $i++)
	{
		for($j = 1; $j <= length($seq1); $j++)
		{
			#We compute the diagonal score using the information of left and up bases
			$letter1 = substr($seq1, $j-1, 1);
			$letter2 = substr($seq2, $i-1, 1);
			$diagonal_score = $matrix[$i-1][$j-1]{score} + $score_matrix{$letter1}{$letter2};
			#We also compute the up and left scores using the gap value defined
			$up_score   = $matrix[$i-1][$j]{score} + $gap;
			$left_score = $matrix[$i][$j-1]{score} + $gap;
			#Based on these three scores, we can chose the highest value and the associated pointer
			if($diagonal_score >= $up_score)
			{
				if($diagonal_score >= $left_score) { $matrix[$i][$j]{score} = $diagonal_score; $matrix[$i][$j]{pointer} = "d"; }
				else { $matrix[$i][$j]{score} = $left_score; $matrix[$i][$j]{pointer} = "l"; }
			}
			else
			{
				if($up_score >= $left_score) { $matrix[$i][$j]{score} = $up_score; $matrix[$i][$j]{pointer} = "u"; }
				else { $matrix[$i][$j]{score} = $left_score; $matrix[$i][$j]{pointer} = "l"; }
			}
		}
	}

	#Defining the needed values to read and define the alignment, starting from the end of the matrix
	$j = length($seq1);
	$i = length($seq2);
	#Reading the complete matrix
	while(1)
	{
		#Stopping if the we reach the first cell of the matrix
		if($matrix[$i][$j]{pointer} eq "none") { last; }
		#using the matrix defined containing a hash (value, and pointer), we read
		#the complete alignment easily
		if($matrix[$i][$j]{pointer} eq "d")
		{
			$align1 .= substr($seq1, $j-1, 1);
			$align2 .= substr($seq2, $i-1, 1);
			$i--;
			$j--;
		}
		elsif($matrix[$i][$j]{pointer} eq "l")
		{
			$align1 .= substr($seq1, $j-1, 1);
			$align2 .= "-";
			$j--;
		}
		elsif($matrix[$i][$j]{pointer} eq "u")
		{
			$align1 .= "-";
			$align2 .= substr($seq2, $i-1, 1);
			$i--;
		}
	}
	#Then, finally, we return the alignment of both reads
	$align1 = reverse $align1;
	$align2 = reverse $align2;
	return($align1, $align2);
}
#End of function
###############################################################################

1;
