########################################################################################################
# PACKAGE NAME: CFunctions
#    FUNCTIONS:	- c_compare (comparison of two sequences given as input fastly, count differences).
#								- c_compare_hm (comparison of two sequences given as input fastly, count differences,
#								but ignoring homopolymer differences of the two sequences).
#		      			- c_compare_len (comparison of two sequences given as input fastly, count differences,
#								but ignoring uncommon parts of the two sequences at the beginning).
#								- c_compare_hm_len (comparison of two sequences given as input fastly, count differences,
#								but ignoring homopolymer differences of the two sequences, and uncommon parts of the two
#								sequences at the beginning).
#
#       AUTHOR:  Sebastien TERRAT <sebastien.terrat@inra.fr> and Corentin Journay
#         TEAM:  UMR 1347 INRA/uB/AgroSup - BioCom Team
#     RELEASED:  30/10/2018
########################################################################################################

package PerlLib::CFunctions;

use strict;
use warnings;

use Inline C => <<'END_OF_C';

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*Function c_compare is based on the comparison of two sequences given as input fastly. This
program takes also the number of differences tolerated (int max) based on the user choices.*/
int c_compare(SV* str_a, SV* str_b,int max)
{
	int diff = 0;
	STRLEN len_a, len_b, len, i;
	char *a, *b;
	char letter;
	//SvPV() function is used to convert the PERL scalar to a string in the C program
	a = SvPV( str_a, len_a );
	b = SvPV( str_b, len_b );
	len = len_a;
	for(i=0; i<len; i++)
	{
		if(a[i]!=b[i]){diff++;}
		if(diff>max){return diff;}
	}
	return diff;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*Function c_compare_len ignoring uncommon parts of the two sequences given as input but taking into
account homopolymer differences. This program takes also the number of differences tolerated (int max)
based on the user choices.*/
int c_compare_len( SV* str_a, SV* str_b, int max )
{
	int diff = 0;
	int verif = 0;
	STRLEN len_a, len_b, len, i;
	//Definition of pointers
	char *a, *b;
	char letter;
	//SvPV() function is used to convert the PERL scalar to a string in the C program
	a = SvPV( str_a, len_a );
	b = SvPV( str_b, len_b );
	len = len_a < len_b ? len_a : len_b;

	for( i = 0; i < len; i++ )
	{
		if ((a[i] == 'A'||a[i] == 'C'||a[i] == 'G'||a[i] == 'T')
		&& (b[i] == 'A'||b[i] == 'C'||b[i] == 'G'||b[i] == 'T')
		&& (verif == 0))
		{
			verif = 1;
		}
		if((a[i] != b[i])&&(verif == 1))
		{
			diff++;
		}
		if(diff > max) return diff;
	}
	//Needed verification as we wait for the start of the two compared sequences
	if (verif == 0) { diff = len; }
	return diff;
}
/*End of function*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*Function c_compare_hm ignoring homopolymer differences of the two sequences given as input. This
program takes also the number of differences tolerated (int max) based on the user choices.*/
int c_compare_hm( SV* str_a, SV* str_b, int max )
{
	int diff = 0;
	int homopolymer_a = 0;
	int homopolymer_b = 0;
	int x = 0;
	STRLEN len_a, len_b, len, i;
	//Definition of pointers
	char *a, *b;
	char letter;
	//SvPV() function is used to convert the PERL scalar to a string in the C program
	a = SvPV( str_a, len_a );
	b = SvPV( str_b, len_b );
	len = len_a < len_b ? len_a : len_b;
	for( i = 0; i < len; i++ )
	{
		if(a[i] != b[i])
		{
			x = 0;
			homopolymer_a = 0;
			if(a[i]=='.'||a[i]=='-') {letter = b[i];}
			else {letter = a[i];}
			//printf("%c ", letter);
			while ((letter == a[i-x]||a[i-x]=='.'||a[i-x]=='-')&& i-x >= 0)
			{
				if (letter == a[i-x]) { homopolymer_a++; }
				x++;
			}
			x = 1;
			while ((letter == a[i+x]||a[i+x]=='.'||a[i+x]=='-')&& i+x <= len)
			{
				if (letter == a[i+x]) { homopolymer_a++; }
				x++;
			}
			//printf("%d ", homopolymer_a);

			x = 0;
			homopolymer_b = 0;
			if(b[i]=='.'||b[i]=='-') { letter = a[i]; }
			else {letter = b[i];}
			while ((letter == b[i-x]||b[i-x]=='.'||b[i-x]=='-')&& i-x >= 0)
			{
				if (letter ==b[i-x]) { homopolymer_b++; }
				x++;
			}
			x = 1;
			while ((letter == b[i+x]||b[i+x]=='.'||b[i+x]=='-')&& i+x <= len)
			{
				if (letter == b[i+x]) { homopolymer_b++; }
				x++;
			}
			if (homopolymer_a > 2 || homopolymer_b > 2)
			{
				if (((homopolymer_a - homopolymer_b == 1)&&(a[i]==b[i]))||((homopolymer_b - homopolymer_a == 1)&&(a[i]==b[i]))
				||(a[i]=='.'||a[i]=='-'||b[i]=='.'||b[i]=='-')) {diff = diff;}
				else { diff++; }
			}
			else { diff++; }
		}
		if(diff > max) return diff;
	}
	return diff;
}
/*End of function*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*Function c_compare_hm_len ignoring homopolymer differences and also uncommon parts of the two
sequences given as input. This program takes also the number of differences tolerated (int max) based on
the user choices.*/
int c_compare_hm_len( SV* str_a, SV* str_b, int max )
{
	int diff = 0;
	int verif = 0;
	int homopolymer_a = 0;
	int homopolymer_b = 0;
	int x = 0;
	STRLEN len_a, len_b, len, i;
	//Definition of pointers
	char *a, *b;
	char letter;
	//SvPV() function is used to convert the PERL scalar to a string in the C program
	a = SvPV( str_a, len_a );
	b = SvPV( str_b, len_b );
	len = len_a < len_b ? len_a : len_b;

	for( i = 0; i < len; i++ )
	{
		if ((a[i] == 'A'||a[i] == 'C'||a[i] == 'G'||a[i] == 'T')
		&& (b[i] == 'A'||b[i] == 'C'||b[i] == 'G'||b[i] == 'T')
		&& (verif == 0))
		{
			verif=1;
		}
		if((a[i] != b[i])&&(verif == 1))
		{
			x = 0;
			homopolymer_a = 0;
			if(a[i]=='.'||a[i]=='-') { letter = b[i]; }
			else {letter = a[i];}
			//printf("%c ", letter);
			while ((letter == a[i-x]||a[i-x]=='.'||a[i-x]=='-')&& i-x >= 0)
			{
				if (letter ==a[i-x]) { homopolymer_a++; }
				x++;
			}
			x = 1;
			while ((letter == a[i+x]||a[i+x]=='.'||a[i+x]=='-')&& i+x <= len)
			{
				if (letter ==a[i+x]) { homopolymer_a++; }
				x++;
			}
			//printf("%d ", homopolymer_a);
			x = 0;
			homopolymer_b = 0;
			if(b[i]=='.'||b[i]=='-') { letter = a[i]; }
			else {letter = b[i];}
			//printf("%c ", letter);
			while ((letter == b[i-x]||b[i-x]=='.'||b[i-x]=='-')&& i-x >= 0)
			{
				if (letter ==b[i-x]) { homopolymer_b++; }
				x++;
			}
			x = 1;
			while ((letter == b[i+x]||b[i+x]=='.'||b[i+x]=='-')&& i+x <= len)
			{
				if (letter ==b[i+x]) { homopolymer_b++; }
				x++;
			}
			if (homopolymer_a > 2 || homopolymer_b > 2)
			{
				if (((homopolymer_a - homopolymer_b == 1)&&(a[i]==b[i]))||((homopolymer_b - homopolymer_a == 1)&&(a[i]==b[i]))
				||(a[i]=='.'||a[i]=='-'||b[i]=='.'||b[i]=='-')) {diff = diff;}
				else { diff++; }
			}
			else { diff++; }
		}
		if(diff > max) return diff;
	}
	//Needed verification as we wait for the start of the two compared sequences
	if (verif == 0) { diff = len; }
	return diff;
}
/*End of function*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
END_OF_C
1;
