###############################################################################
#POD Part
###############################################################################

=head1 NAME

package I<CheckChoicesFolders> - library used to manage, check, and treat user
choices, given as options when launching the main program, or by answer(s) when
the treatment is realized.

=head1 LIBRARY ORGANIZATION AND FUNCTIONS

	use CheckChoicesFolders;
	TreatInputArgumentsDB();
	TreatInputArgumentsClust();
	CatchUserChoices();
	DefineDatabaseDirectory(choices in array);
	CleanFolders();

=head1 DESCRIPTION

This module was defined and used to manage, check, and treat user choices,
given as options when launching the main program, or by answer(s) when the
treatment is done (automatically or manually).
The documentation was realized with B<Pod>.

=head2 FUNCTIONS

=over 4

=item B<TreatInputArgumentsDB()>

Check and treat the input user options in the command line for the
Database_Format program. For each option, this function checks if the given
file(s) exists or not, and store the details in the %Input_args hash. Moreover,
it checks also if the good number of options have been given by the user.

=item B<TreatInputArgumentsClust()>

Check and treat the input user options in the command line for the
Database_Clustering program. For each option, this function checks if the given
file(s) exists or not, and store the details in the %Input_args hash. Moreover,
it checks also if the good number of options have been given by the user.

=item B<CatchUserChoices()>

Catch all user choices based on his answers when the program asked for
information (database name, single-singleton treatment, percentage of
similarity, taxonomy chosen level and type of organism).

=item B<DefineDatabaseDirectory()>

Check and define the needed folder to store the Database (chosen by the user).
More precisely, define the directory where the database element will be stored
automatically by the program and check the existence of previous versions

=item B<CleanFolders()>

Print the complete help for the Database_Clustering program. Clean and create
if needed the Temporary_files and Result folders. For both folders, the function
will delete all files in these folders before starting the analysis. We finally
return the Result folder name to the main program.

=back

=head1 AUTHORS

Sebastien TERRAT <sebastien.terrat@inra.fr>
Corentin Journay

=head1 TEAM AND NOTES

UMR 1347 INRA/uB/AgroSup - BioCom Team
RELEASED:  30/10/2018

=cut

###############################################################################
#FUNCTIONS AND PROGRAMS
###############################################################################

# Name of the package defined for the main program
package PerlLib::CheckChoicesFolders;
use strict;
use warnings;

###############################################################################
#FONCTION:		TreatInputArgumentsDB
#USAGE:				TreatInputArgumentsDB()
#ARGUMENTS:		None
#DETAILS:			Check and catch all given input arguments by the user when the
#		command line is written. For each option, this function checks
#		if the given file exists or not, and store the details in the
#		%Input_args  hash. Moreover, it checks also if the good number
#		of options have been given by the user.
#
#RELEASED: 05/10/2018
###############################################################################
sub TreatInputArgumentsDB
{
	#Defined variables
	my %Input_args = ();
	my $i = 0;
	my $msg_clustering = "==> The given clustering filename doesn't exists.\n";
	my $msg_taxonomy = "==> The given taxonomy filename doesn't exists.\n";
	my $msg_derep = "==> The given dereplication filename doesn't exists.\n";
	my $msg_seq = "==> The given FASTA sequence filename doesn't exists.\n";
	my $msg_cpu = "==> The given number of CPUs or cores was not an integer or below 0, the program will use the default value.\n";

	#Test to check the good number of input arguments
	if ($#ARGV < 7 ) { PerlLib::Help::HelpDB(); exit(0); }
	#For all input arguments, we checked and organize the given data.
	for($i=0;$i<=$#ARGV;$i++)
	{
		if($ARGV[$i] eq "-clust")
		{
			if(!exists($ARGV[$i+1])) { die "$msg_clustering"; }
			if(!-e $ARGV[$i+1]) { die "$msg_clustering"; }
			else { $Input_args{$ARGV[$i]} = $ARGV[$i+1]; }
		}
		if($ARGV[$i] eq "-taxo")
		{
			if(!exists($ARGV[$i+1])) { die "$msg_taxonomy"; }
			if(!-e $ARGV[$i+1]) { die "$msg_taxonomy"; }
			else { $Input_args{$ARGV[$i]} = $ARGV[$i+1]; }
		}
		if($ARGV[$i] eq "-derep")
		{
			if(!exists($ARGV[$i+1])) { die "$msg_derep"; }
			if(!-e $ARGV[$i+1]) { die "$msg_derep"; }
			else { $Input_args{$ARGV[$i]} = $ARGV[$i+1]; }
		}
		if($ARGV[$i] eq "-seq")
		{
			if(!exists($ARGV[$i+1])) { die "$msg_seq"; }
			if(!-e $ARGV[$i+1]) { die "$msg_seq"; }
			else { $Input_args{$ARGV[$i]} = $ARGV[$i+1]; }
		}
		if($ARGV[$i] eq "-cp")
		{
			$ARGV[$i+1] = $ARGV[$i+1] + 0;
			if($ARGV[$i+1] == 0) { print "$msg_cpu"; }
			else { $Input_args{$ARGV[$i]} = $ARGV[$i+1]; }
		}
	}

	#Defining defaut values
	if(!exists($Input_args{"-cp"})) { $Input_args{"-cp"} = 1; }
	#We finally return all data organized in the hash table
	return(\%Input_args);
}
#End of function
###############################################################################

###############################################################################
#FONCTION:		TreatInputArgumentsClust
#USAGE:				TreatInputArgumentsClust()
#ARGUMENTS:		None
#DETAILS:			Check and catch all given input arguments by the user when the
#		command line is written. For each option, this function checks
#		if the given file exists or not, and store the details in the
#		%Input_args  hash. Moreover, it checks also if the good number
#		of options have been given by the user.
#
#RELEASED: 05/10/2018
###############################################################################
sub TreatInputArgumentsClust
{
	#Defined variables
	my %Input_args = ();
	my $i = 0;
	my $msg_db = "==> The given database folder doesn't exists.\n";
	my $msg_threshold = "==> The given clustering threshold is out of range [1-100]. Please try again.\n";
	my $msg_clust = "==> The given clustering type is not clear [infern/needle]. Please try again.\n";
	my $msg_seq = "==> The given FASTA sequence filename or folder doesn't exists.\n";
	my $msg_cpu = "==> The given number of CPUs or cores was not an integer or below 0, the program will use the default value.\n";
	my $msg_o = "==>The given output folder contain not valid characters, the program is stopped. Please try again.\n";
	my $msg_o_ok = "==>The given output folder does not exists, the program will use the default value.\n";

	#Test to check the good number of input arguments
	if ($#ARGV < 7 ) { PerlLib::Help::HelpClust(); exit(0); }
	#For all input arguments, we checked and organize the given data.
	for($i=0;$i<=$#ARGV;$i++)
	{
		if($ARGV[$i] eq "-db")
		{
			if($ARGV[$i+1] =~ m/$\//) { $ARGV[$i+1] =~ s/\//\//; }
			else { $ARGV[$i+1] = $ARGV[$i+1]."/"; }
			if(!exists($ARGV[$i+1])) { die "$msg_db"; }
			if(!-e $ARGV[$i+1]) { die "$msg_db"; }
			else { $Input_args{$ARGV[$i]} = $ARGV[$i+1]; }
		}
		if($ARGV[$i] eq "-sq")
		{
			if(!exists($ARGV[$i+1])) { die "$msg_seq"; }
			if(!-e $ARGV[$i+1]) { die "$msg_seq"; }
			else { $Input_args{$ARGV[$i]} = $ARGV[$i+1]; }
		}
		if($ARGV[$i] eq "-th")
		{
			$ARGV[$i+1] = $ARGV[$i+1] + 0;
			if(($ARGV[$i+1] <= 0)||($ARGV[$i+1] > 100)) { die "$msg_threshold"; }
			else { $Input_args{$ARGV[$i]} = $ARGV[$i+1]; }
		}
		if($ARGV[$i] eq "-cl")
		{
			chomp($ARGV[$i+1]);
			if((uc($ARGV[$i+1]) ne "INFERN")&&(uc($ARGV[$i+1]) ne "NEEDLE")) { die "$msg_clust"; }
			else { $Input_args{$ARGV[$i]} = $ARGV[$i+1]; }
		}
		if($ARGV[$i] eq "-cp")
		{
			$ARGV[$i+1] = $ARGV[$i+1] + 0;
			if($ARGV[$i+1] == 0) { print "$msg_cpu"; }
			else { $Input_args{$ARGV[$i]} = $ARGV[$i+1]; }
		}
		if($ARGV[$i] eq "-o")
		{
			if($ARGV[$i+1] =~ m/[^\w\_\-\.]/) { die "$msg_o"; }
			if(!-e $ARGV[$i+1])
			{
				print "$msg_o_ok";
				$Input_args{"-o"} = "Results/";
			}
			else
			{
				if($ARGV[$i+1] =~ m/$\//) { $ARGV[$i+1] =~ s/\//\//; }
				else { $ARGV[$i+1] = $ARGV[$i+1]."/"; }
				$Input_args{$ARGV[$i]} = $ARGV[$i+1];
			}
		}
		#Optional arguments given as input by the user
		if($ARGV[$i] eq "-h") { $Input_args{"-h"} = ""; }
		if($ARGV[$i] eq "-c") { $Input_args{"-c"} = ""; }
		if($ARGV[$i] eq "-hc") { $Input_args{"-hc"} = ""; }
		if($ARGV[$i] eq "-ad") { $Input_args{"-ad"} = ""; }
	}

	#Defining defaut values
	if(!exists($Input_args{"-cp"})) { $Input_args{"-cp"} = 1; }
	if(!exists($Input_args{"-o"})) { $Input_args{"-o"} = "Results/"; }

	#We finally return all data organized in the hash table
	return(\%Input_args);
}
#End of function
###############################################################################

###############################################################################
#FUNCTION:	CatchUserChoices
#USAGE:			CatchUserChoices()
#ARGUMENTS:	None
#DETAILS:		Ask questions to the user to obtain needed information before
#		database definition.
#RELEASED: 09/10/2018
###############################################################################
sub CatchUserChoices
{
	#Defined variables
	my $name = "";
	my $choice_SS = "";
	my $clust_level = 0;
	my $taxo_level = 0;
	my $model = "";
	my @choices = ();

	# User choice to define the name of the database and treatment and checking
	printf "Please choose the name of the database (short name whithout spaces or special characters).\n==> ";
	$name = <STDIN>;
	chomp($name);
	if($name =~ m/[^A-Za-z0-9\_\-]+/) { die "The chosen name contains not wanted characters. Please try again.\n"; }

	# User choice to define if the OTUs with 1 read (single-singletons) are kept or not
	printf "Will you want to keep the single-singletons in the created database (yes-no) ?\n==> ";
	$choice_SS = <STDIN>;
	chomp($choice_SS);
	$choice_SS = uc($choice_SS);
	if((!($choice_SS eq "YES"))&&(!($choice_SS eq "Y"))&&(!($choice_SS eq "NO"))&&(!($choice_SS eq "N")))
	{
		die "Your choice is not clear. Please answer yes, no, Y or N\n";
	}

	# Needed information to efficiently create the database (similarity clustering level)
	printf "Please indicate the similarity level defined for the clustering step of your samples ?(default: 95)[1-100].\n==> ";
	$clust_level = <STDIN>;
	if($clust_level eq "\n") {$clust_level = 95; }
	else { chomp($clust_level); }
	$clust_level = $clust_level + 0;
	if(($clust_level <= 0)||($clust_level > 100)) { die "The given similarity level is below the out of range [1-100]. Please try again.\n"; }

	# Needed information to efficiently create the database (taxonomic confidence/similarity level)
	printf "Please indicate the level used to define the sequence taxonomy of your samples ?(default: 80)[50-100].\n==> ";
	$taxo_level = <STDIN>;
	if($taxo_level eq "\n") {$taxo_level = 80; }
	else {chomp($taxo_level); }
	$taxo_level = $taxo_level + 0;
	if(($taxo_level <= 49)||($taxo_level > 100)) { die "The given taxonomic similarity level is out of the defined range [50-100]. Please try again.\n"; }

	# Needed information to efficiently create the database (taxonomic confidence/similarity level)
	printf "Please indicate the targeted organism (to define the potential model used in the global alignment) [bacteria/fungi/archaea/algae]\n==> ";
	$model = <STDIN>;
	chomp($model);
	$model = uc($model);
	if((!($model eq "BACTERIA"))&&(!($model eq "FUNGI"))&&(!($model eq "ARCHAEA"))&&(!($model eq "ALGAE")))
	{
		die "Your choice is not available. Please choose between bacteria, fungi, archaea or algae.\n";
	}
	@choices = ($name, $choice_SS, $clust_level, $taxo_level, $model);
	return(\@choices);
}
#End of function
###############################################################################

###############################################################################
#FUNCTION:	DefineDatabaseDirectory
#USAGE:			DefineDatabaseDirectory(ref_array of choices)
#ARGUMENTS:	ref_array of choices
#DETAILS:		Define the directory where the database element will be
#		stored automatically by the program and check the existence of
#		previous versions
#RELEASED: 09/10/2018
###############################################################################
sub DefineDatabaseDirectory
{
	#Defined variables
	my @choices = @{$_[0]};
	my $folder = "";
	my $choice_SS = "";
	#Based on the user choice, we indicate the define $choice_SS ("G" for
	#global and "F" for filtered).
	if($choices[1] eq "YES") { $choice_SS = "G"; }
	else { $choice_SS = "F"; }
	#We define the database folder name
	$folder = "DB_".$choices[0]."_".$choices[4]."_C".$choices[2]."_T".$choices[3]."_".$choice_SS;
	if(!-d $folder) { mkdir($folder); }

	if(-e $folder."/database_".$choices[0]."_".$choices[4]."_C".$choices[2]."_T".$choices[3]."_".$choice_SS)
	{
		die "The database file database_".$choices[0]."_".$choices[4]."_C".$choices[2]."_T".$choices[3]."_".$choice_SS." already exists.
Please check and relaunch the program.\n";
	}
	if(-e $folder."/dbreads_".$choices[0]."_".$choices[4]."_C".$choices[2]."_T".$choices[3]."_".$choice_SS)
	{
		die "The dbreads file dbreads_".$choices[0]."_".$choices[4]."_C".$choices[2]."_T".$choices[3]."_".$choice_SS." already exists.
Please check and relaunch the program.\n";
	}
	#We finally return the name of the created folder to store and use it
	#in further steps.
	return($folder);
}
#End of function
###############################################################################

###############################################################################
#FUNCTION:	CleanFolders
#USAGE:			CleanFolders()
#ARGUMENTS:	None
#DETAILS:		Clean and create if needed the Temporary_files and Result folders
#		For both folders, the function will	delete all files in these
#		folders before starting the analysis. We finally return the result
#		folder name to the main program.
#
#RELEASED: 11/2018
###############################################################################
sub CleanFolders
{
	#Defined variables
	my $chosen_output_folder = $_[0];
	my $file = "";
	#my $result_folder = "Results"

	#We cleaned the temporary directory from files
	if((-d "Temporary_files")&&(-e "Temporary_files"))
	{
		opendir (DIR, "Temporary_files/") || die "can't open the directory Temporary_files/: $!";
		while (defined($file = readdir(DIR))) { unlink("Temporary_files/$file"); }
		closedir(DIR);
	}
	#Or we create it if needed
	else  {mkdir("Temporary_files"); }

	#We cleaned the result directory from files
	if((-d "$chosen_output_folder")&&(-e "$chosen_output_folder"))
	{
		opendir (DIR, "$chosen_output_folder") || die "can't open the directory $chosen_output_folder/: $!";
		while (defined($file = readdir(DIR))) { unlink("$chosen_output_folder"."$file"); }
		closedir(DIR);
	}
	#Or we create it if needed
	else { mkdir("$chosen_output_folder"); }
	#We finally return the result folder name to the main program.
	print "# Temporary and $chosen_output_folder folders checked and cleaned\n";
	return($chosen_output_folder);
}
#End of function
#############################################################################

1;
