###############################################################################
#POD Part
###############################################################################

=head1 NAME

package I<CheckFormats> - library used to check the structure of all input
file(s) given by the user of both programs (Database_Format, and
Database_Clustering). Moreover, inside each file, the structure of each line is
also checked.

=head1 LIBRARY ORGANIZATION AND FUNCTIONS

	use CheckFormats;
	CheckFileFormats(\%Input_args);
	CheckReadStructure();
	CheckSeqFile();
	CheckClustFile();
	CheckDerepFile();
	CheckTaxoFile();
	CheckDBFolder();
	CheckDBFile();

=head1 DESCRIPTION

This module was defined to check automatically all given input files (fasta,
aligned reads, taxonomy, etc.) regarding their structure. Indeed, these files
will be treated by the main programs, and B<MUST> have good formats to avoid
bugs and problems during the treatment. It can take some minutes to treat it.
The documentation was realized with B<Pod>.

=head2 FUNCTIONS

=over 4

=item B<CheckFileFormats()>

Choose the good function to check the format of the file. Takes as input the
reference hash table with all input filenames ($_[0]). Based on the key in
the hash table associated to each filename, the function choose the good
function to check and treat the entire file.

=item B<CheckReadStructure()>

This function check the structure of the sequence ID, based on regular
expressions. It verifies the presence of all elements, and stop the program if
needed with an error message for the user. Takes as input three elements: the
origin of the sequence ID, the ID of the sequence and the filename given as
input.

=item B<CheckSeqFile()>

This function check the structure of the sequence file, based on regular
expressions. It verifies the presence and the structure of all elements, and
stop the program if needed with an error message for the user. It takes as input
the origin of the filename (DB or Clust), and the filename currently treated.

=item B<CheckClustFile()>

This function check the structure of the clustering file, based on regular
expressions. It verifies the presence and the structure of all elements, and
stop the program if needed with an error message for the user. It takes as input
the origin of the filename (DB or Clust), and the filename currently treated.

=item B<CheckDerepFile()>

This function check the structure of the dereplicated file, based on regular
expressions. It verifies the presence and the structure of all elements, and
stop the program if needed with an error message for the user. It takes as input
the origin of the filename (DB or Clust), and the filename currently treated.

=item B<CheckTaxoFile()>

This function check the structure of the taxonomic file, based on regular
expressions. It verifies the presence and the structure of all elements, and
stop the program if needed with an error message for the user. It takes as input
the name of the currently treated file.

=item B<CheckDBFolder()>

This function check the different files found in the folder given as input by
the user and the structure of each found file. It takes as input only the name
of the folder given by the user.

=item B<CheckDBFile()>

This function check all lines from the database file found in the given folder
name by the user based on regular expressions. It verifies the presence and the
structure of all elements, and stop the program if needed with an error message
for the user. It takes as input the database filename found in the folder.

=back

=head1 AUTHORS

Sebastien TERRAT <sebastien.terrat@inra.fr> and Corentin Journay

=head1 TEAM AND NOTES

UMR 1347 INRA/uB/AgroSup - BioCom Team
RELEASED:  30/10/2018

=cut

###############################################################################
#FUNCTIONS AND PROGRAMS
###############################################################################

# Name of the package defined for the main program
package PerlLib::CheckFormats;
use strict;
use warnings;

#############################################################################
#FUNCTION:		CheckFileFormats
#USAGE:				CheckFileFormats(\%Input_args)
#ARGUMENTS:		The reference hash table with all input filenames ($_[0])
#DETAILS:		Based on the key in the hash table associated to each
#			filename, the function choose the good function to treat
#			the entire file.
#
#RELEASED: 05/10/2018
#############################################################################
sub CheckFileFormats
{
	#Defined variables
	my %Input_args = %{$_[0]};
	my $key = "";
	#Here, we launch the good file checking
	foreach $key (keys(%Input_args))
	{
		#Regarding the options for the Database Definition Program
		if($key eq "-clust")	{ &CheckClustFile("-clust", $Input_args{$key}); }
		if($key eq "-taxo")	{ &CheckTaxoFile($Input_args{$key});}
		if($key eq "-derep")	{ &CheckDerepFile("-derep", $Input_args{$key}); }
		if($key eq "-seq")	{ &CheckSeqFile("-seq", $Input_args{$key}); }
		#Regarding the options for the Clustering Program
		if($key eq "-sq")		{ &CheckSeqFile("-sq", $Input_args{$key}); }
		if($key eq "-db")		{ &CheckDBFolder($Input_args{$key}); }
	}
}
#End of function
#############################################################################

#############################################################################
#FUNCTION:		CheckReadStructure
#USAGE:				CheckReadStructure(ARGUMENTS)
#ARGUMENTS:		- The origin of the sequence ID ($_[0]).
#							- The ID of a sequence ($_[1]).
#							- The filename given as input ($_[2]).
#DETAILS:		This function check the structure of the sequence ID, based
#			on regular expressions. It verifies the presence of all
#			elements, and stop the program if needed with an error
#			message for the user.
#
#RELEASED: 05/10/2018
#############################################################################
sub CheckReadStructure
{
	#Defined variables
	my $opt_origin = $_[0];
	my $ID_read = $_[1];
	my @tmp = ();
	my $msg_read = "==> The sequence IDs are not formatted correctly in the
$_[2] file, please check it.\n ";
	my $msg_db = "==> The sequence IDs of the database file are not formatted correctly
 in the $_[2] file, please check it.\n ";

	#We check the format of all sequences descriptive lines
	@tmp = split("_", $ID_read);
	#Here, we check for the format regarding the database program
	if($opt_origin eq "-seq")
	{
		if(!($tmp[0] =~ m/^\d*\|[A-Za-z0-9]*/)) { die $msg_read; }
	}
	#Here, we check for the format regarding the clustering program
	elsif($opt_origin eq "-sq")
	{
		if(!($tmp[0] =~ m/^[A-Za-z0-9]*/)) { die $msg_read; }
	}
	#Here, we check for the format regarding the clustering program
	#for the database itself
	elsif($opt_origin eq "dbreads")
	{
		if(!($tmp[0] =~ m/^OTU[0-9]*/)) { die $msg_db; }
	}
	#Avoid some checking for the database
	if($opt_origin ne "dbreads")
	{
		if(!($tmp[1] =~ m/\d*/)) { die $msg_read; }
		if(!($tmp[2] =~ m/length=\d*/)) { die $msg_read; }
	}
}

#############################################################################
#FUNCTION:		CheckSeqFile
#USAGE:				CheckSeqFile(ARGUMENTS)
#ARGUMENTS:		- The origin or the filename (DB or Clust program) ($_[0])
#							- The filename currently treated ($_[1])
#DETAILS:		This function check the structure of the sequence file,
#			based on regular expressions. It verifies the presence and
#			the structure of all elements, and stop the program if
#			needed with an error message for the user.
#
#RELEASED: 05/10/2018
#############################################################################
sub CheckSeqFile
{
	#Defined variables
	my $opt_origin = $_[0];
	my $filename = $_[1];
	my $char = "";
	my @tmp = ();
	my $msg_seq = "==>The FASTA file does not contain ID sequences or
FASTA sequences with the good format, please check your file\n";
	my $msg_db = "==>The FASTA file of the database does not contain
ID sequences or FASTA sequences with the good format, please check your file\n";

	#Opening of the sequence file
	open(F_SEQ, $filename) || die "Can't open $filename\n";
	#For each line found in the sequence file
	while(<F_SEQ>)
	{
		chomp($_);
		#If the reading line starts with the character ">"
		if($_ =~ m/^>/)
		{
			$_ =~ s/^>//;
			#We check the format of all sequences descriptive lines
			&CheckReadStructure($opt_origin, $_, $filename);
		}
		#If it is the sequence itself
		else
		{
			#We check the format of all sequences lines
			@tmp = split("", $_);
			foreach $char (@tmp)
			{
				if(!($char =~ m/[ACGTacgt]/))
				{
					if($opt_origin eq "dbreads") { die $msg_db; }
					else { die $msg_seq; }
				}
			}
		}
	}
	close(F_SEQ);
	print "# Sequence file $filename format checked and passed\n";
}
#End of function
#############################################################################

#############################################################################
#FUNCTION:		CheckClustFile
#USAGE:				CheckClustFile(ARGUMENTS)
#ARGUMENTS:		- The origin or the filename (DB or Clust program) ($_[0])
#							- The filename currently treated ($_[1])
#DETAILS:			This function check the structure of the clustering file,
#			based on regular expressions. It verifies the presence and
#			the structure of all elements, and stop the program if
#			needed with an error message for the user.
#
#RELEASED: 05/10/2018
#############################################################################
sub CheckClustFile
{
	#Defined variables
	my $opt_origin = $_[0];
	my $filename = $_[1];
	my $element = "";
	my @tmp = ();
	my $msg_clust_otu = "==>The clustering file does not contain OTU IDs,
please check your file\n";
	my $msg_nbreads_otu = "==>The clustering file does not contain the number
of sequences in each OTU, please check your file\n";

	#Opening of the clustering file
	open(F_CLUST, $filename) || die "Can't open $filename\n";
	#For each line of the clustering file
	while(<F_CLUST>)
	{
		chomp($_);
		#If the reading line starts with a number
		if($_ =~ m/^\d/)
		{
			#We check the format of all sequences descriptive lines
			@tmp = split("\t", $_);
			if(!($tmp[0] =~ m/\d*/)) { die $msg_clust_otu; }
			if(!($tmp[2] =~ m/\d*/)) { die $msg_nbreads_otu; }
			@tmp = split(" ", $tmp[3]);
			foreach $element (@tmp) { &CheckReadStructure($opt_origin, $element, $filename); }
		}
	}
	close(F_CLUST);
	print "# Clustering file format checked and passed\n";
}
#End of function
#############################################################################

#############################################################################
#FUNCTION:	CheckDerepFile
#USAGE:			CheckDerepFile(ARGUMENTS)
#ARGUMENTS:	- The origin or the filename (DB or Clust program) ($_[0])
#						- The filename currently treated ($_[0])
#DETAILS:		This function check the structure of the dereplicated file,
#			based on regular expressions. It verifies the presence and
#			the structure of all elements, and stop the program if
#			needed with an error message for the user.
#
#RELEASED: 05/10/2018
#############################################################################
sub CheckDerepFile
{
	#Defined variables
	my $opt_origin = $_[0];
	my $filename = $_[1];
	my $element = "";
	my @tmp = ();
	#Opening of the dereplication file
	open(F_DEREP, $filename) || die "Can't open $filename\n";
	#For each line, it checks all elements after splitting
	while(<F_DEREP>)
	{
		chomp($_);
		@tmp = split("\t", $_);
		&CheckReadStructure($opt_origin, $tmp[0], $filename);
		@tmp = split(" ", $tmp[1]);
		foreach $element (@tmp) { &CheckReadStructure($opt_origin, $element, $filename); }
	}
	close(F_DEREP);
	print "# Dereplication file format checked and passed\n";
}
#End of function
#############################################################################

#############################################################################
#FUNCTION:		CheckTaxoFile
#USAGE:				CheckTaxoFile($filename)
#ARGUMENTS:		- The filename currently treated ($_[0])
#DETAILS:		This function check the structure of the taxonomic file,
#			based on regular expressions. It verifies the presence and
#			the structure of all elements, and stop the program if
#			needed with an error message for the user.
#
#RELEASED: 05/10/2018
#############################################################################
sub CheckTaxoFile
{
	#Defined variables
	my $filename = $_[0];
	my $element = "";
	my $i = 0;
	my @tmp = ();
	my @tmp2 = ();
	my $msg_taxo = "==>The taxonomic file does not contain 5 taxonomic levels and OTU IDs, in the same line,
please check your taxonomic file\n";
	my $msg_taxo_details = "==>The taxonomic file does not contain available taxonomic names in line,
please check your taxonomic file\n";
	#Opening of the dereplication file
	open(F_TAXO, $filename) || die "Can't open $filename\n";
	#For each line, it checks all elements after splitting
	while(<F_TAXO>)
	{
		chomp($_);
		@tmp = split("\t", $_);
		#Check if the we have the good start of line (OTU and/or the OTU ID)
		if(($tmp[0] =~ m/(^\d+)/)||($tmp[0] =~ m/(^OTU\d+)/))
		{
			#We must have 6 elements for the line (OTU ID and 5 taxo levels)
			if ($#tmp < 5) { die "$msg_taxo"; }
			#We also check each element
			for($i=1;$i<=$#tmp;$i++)
			{
				@tmp2 = split(" ", $tmp[$i]);
				#We must have at least one name for each taxonomic level
				if($#tmp2 > -1)
				{
					#Then, we check the structure of the name itself
					foreach $element (@tmp2)
					{
						if($#tmp2 == 0)
						{
							if((!($element =~ m/.*\(\d+\)/))||(!($element =~ m/.*/))) { die "$msg_taxo_details"; }
						}
						#If we have several names, we must have the number of associated reads
						else { if(!($element =~ m/.*\(\d+\)/)) { die "$msg_taxo_details"; } }
					}
				}
				else { die "$msg_taxo"; }
			}
		}
	}
	close(F_TAXO);
	print "# Taxonomic file format checked and passed\n";
}
#End of function
#############################################################################

#############################################################################
#FUNCTION:		CheckDBFolder
#USAGE:				CheckDBFolder($foldername)
#ARGUMENTS:		- The database folder given as input by the user ($_[0])
#DETAILS:			This function check the different files found in the folder
#			given as input by the user and the structure of each found
#			file.
#
#RELEASED: 05/10/2018
#############################################################################
sub CheckDBFolder
{
	#Defined variables
	my $DB_folder = $_[0];
	my @filenames = ();
	my $filename = "";
	my $msg_db = "==>The database folder does not exists, please check
the given database folder name.\n";


	#Depending of the folder given as argument, we catch all filenames
	#and we check if the DB exists
	if((-d $DB_folder)&&(-e $DB_folder))
	{
		opendir (DIR, "$DB_folder/") || die "can't open the directory $DB_folder: $!";
		LOOP01:while (defined($filename = readdir(DIR)))
		{
			if ($filename =~ m/^\.\.?$/) { next LOOP01; }
			else { push(@filenames, $filename); }
		}
		closedir(DIR);
	}
	else { die $msg_db; }
	#Then, for each file found
	foreach $filename (@filenames)
	{
		if($filename =~ m/^dbreads_/)
		{
			$filename = $DB_folder.$filename;
			&CheckSeqFile("dbreads", $filename);
		}
		elsif($filename =~ m/^database_/)
		{
			$filename = $DB_folder.$filename;
			&CheckDBFile($filename);
		}
	}
}
#End of function
#############################################################################

#############################################################################
#FUNCTION:		CheckDBFile
#USAGE:				CheckDBFile($filename)
#ARGUMENTS:		- The database file found in the folder given as input
#by the user ($_[0])
#DETAILS:			This function check all lines from the database file found
#			in the given folder name by the user based on regular
#			expressions. It verifies the presence and the structure of
#			all elements, and stop the program if needed with an error
#			message for the user.
#
#RELEASED: 05/10/2018
#############################################################################
sub CheckDBFile
{
	#Defined variables
	my $filename = $_[0];
	my @tmp = ();
	my @tmp2 = ();
	my @tmp3 = ();
	my $element = "";
	my $msg_database = "==>The database file in the given folder is not
coherent, please check the database file.\n";

	#Opening of the database file
	open(F_DB, $filename) || die "Can't open $filename\n";
	#For each line, it checks all elements after splitting
	while(<F_DB>)
	{
		#First, we split the line
		chomp($_);
		@tmp = split("\t", $_);
		#We then check all elements from the database file and their structure
		if(!($tmp[0] =~ m/^OTU[0-9]*/)) { die $msg_database; }
		if(!($tmp[1] =~ m/^[0-9]*/)) { die $msg_database; }
		@tmp2 = split(/\|/, $tmp[2]);
		foreach $element (@tmp2)
		{
			if(!($element =~ m/^[0-9]\.[0-9][0-9]\_[A-Za-z_-]/)) { die $msg_database; }
		}
		@tmp2 = split(/\|/, $tmp[3]);
		foreach $element (@tmp3)
		{
			if(!($element =~ m/^\d*\_\d*/)) { die $msg_database; }
		}
	}
	close(F_DB);
}
#End of function
#############################################################################

1;
