###############################################################################
#POD Part
###############################################################################

=head1 NAME

package I<Clustering> - library used to realize, compile and clean results from
the clustering step.

=head1 LIBRARY ORGANIZATION AND FUNCTIONS

	use Clustering;
	DefineThresholdClust();
	CompileClusteringResult();
	CleaningKeepingOuts();
	ClusteringAgainstDatabase();
	ClusteringDeNovo();

=head1 DESCRIPTION

This module was defined and used to realize, compile and clean results from
the clustering step. This module contains functions to realize the clustering,
both against the database and de novo. But also functions to compile results and
cleaning data.
The documentation was realized with B<Pod>.

=head2 FUNCTIONS

=over 4

=item B<DefineThresholdClust()>

This function is defined to compute the number of differences tolerated into the
comparison between both reads, and also the length of the read studied. It takes
as input the number of gaps between both reads, and also the analyzed read and
the chosen threshold. A output, it returns the length of the read (aligned) and
the number of differences based on th threshold chosen by the user.

=item B<CompileClusteringResult()>

This function is defined to print out the result needed for the definition of
precise result files for the user. It takes several elements as input: the
header of the output file ($_[0]), the sequence ID that need treatment ($_[1]),
the sequence ID from the database with results ($_[2]), the number of gaps found
in the DB read ($_[3]), the sequence from the file that need treatment ($_[4]),
the defined threshold by the user ($_[5]) and the number of differences found by
clustering ($_[6]). It prints results on the raw result file.

=item B<CleaningKeepingOuts()>

This function is defined to delete not needed data stored in two hashes from the
main program, to easily treat needed data that did not match in the database by
a de novo clustering step. It takes as input two references of two hashes: the
sample and the abundance ones. The same cleaned hashes are then given as output.

=item B<ClusteringAgainstDatabase()>

This function is defined to realize the clustering of sequences given by the
user as input against the database chosen by the user. Several elements are
needed as input, but the program give back no output.

=item B<ClusteringDeNovo()>

This function is defined to realize the clustering of sequences given by the
user as input by a de novo approach. Several elements are needed as input, but
the program give back no output.

=back

=head1 AUTHORS

Sebastien TERRAT <sebastien.terrat@inra.fr>
Corentin Journay

=head1 TEAM AND NOTES

UMR 1347 INRA/uB/AgroSup - BioCom Team
RELEASED:  30/10/2018

=cut

###############################################################################
#FUNCTIONS AND PROGRAMS
###############################################################################

# Name of the package defined for the main program
package PerlLib::Clustering;
use strict;
use warnings;

#############################################################################
# FUNCTION:		DefineThresholdClust
# USAGE:			DefineThresholdClust(arguments)
# ARGUMENTS:	- The number of gaps found in the DB read ($_[0])
#			- The sequence from the file that need treatment ($_[1])
#			- The defined threshold by the user ($_[2])
# DETAILS:		This function is defined to compute the number of
#			differences tolerated into the comparison between both
#			reads, and also the length of the read studied.
#
# RELEASED: 10/10/2018
#############################################################################
sub DefineThresholdClust
{
	#Defined variables
	my $gaps_db = $_[0];
	my $seq = $_[1];
	my $given_threshold = $_[2];
	my $length = 0;
	my $count = 0;
	my $threshold = 0;

	#We count into the sequence the number of gaps found
	$count = $seq =~ tr/\-\./\-\./;
	#We defined the length of the shorter read using the gaps
	if($gaps_db <= $count) { $length = length($seq)-$count; }
	else { $length = length($seq)-$gaps_db; }
	#We then define the threshold using theis value
	$threshold = int((((100-$given_threshold)/100)*$length)+0.5);
	#We return these two info: the length and the threshold (nb of bases)
	return($length, $threshold);
}
#End of function
#############################################################################

#############################################################################
# FUNCTION:		CompileClusteringResult
# USAGE:			CompileClusteringResult(arguments)
# ARGUMENTS:	- The header of the output file ($_[0])
#			- The sequence ID from the file that need treatment ($_[1])
#			- The sequence ID from the database with result ($_[2])
#			- The number of gaps found in the DB read ($_[3])
#			- The sequence from the file that need treatment ($_[4])
#			- The defined threshold by the user ($_[5])
#			- The number of differences found by clustering ($_[6])
# DETAILS:		This function is defined to print out the result needed for
#			the definition of precise result files for the user.
#
# RELEASED: 10/10/2018
#############################################################################
sub CompileClusteringResult
{
	#Defined variables
	my $output_header = $_[0];
	my $ID = $_[1];
	my $ID_DB = $_[2];
	my $gaps_otu = $_[3];
	my $seq_read = $_[4];
	my $chosen_threshold = $_[5];
	my $nb_diff_found = $_[6];
	my ($len_seq, $theoric_threshold, $percent_id) = (0,0,0);

	#First, we launch the needed function to compute one last time the length of the read
	($len_seq, $theoric_threshold) = &DefineThresholdClust($gaps_otu, $seq_read, $chosen_threshold);
	#We then define the result found in terms of identity
	$percent_id = ($len_seq - $nb_diff_found)/$len_seq;
	#We finally print this result in the output file
	flock($output_header, 2);
	printf($output_header $ID."\t".sprintf("%.2f",$percent_id*100)."\t".$ID_DB."\n");
	flock($output_header, 8);
	#And return an empty variable to redefine an input variable
	return("");
}
#End of function
#############################################################################

#############################################################################
# FUNCTION:		CleaningKeepingOuts
# USAGE:			CleaningKeepingOuts(arguments)
# ARGUMENTS:	- The reference of the %sample hash ($_[0]),
#			- The reference of the %abundance hash ($_[1])
# DETAILS:		This function is defined to delete not needed data stored
#			in two hashes from the main program, to easily treat needed
#			data that did not match in the database.
#
# RELEASED: 10/10/2018
#############################################################################
sub CleaningKeepingOuts
{
	#Defined variables
	my %sample = %{$_[0]};
	my %abundance = %{$_[1]};
	my %IDs_out = ();
	my $key = "";

	#Here, we open and catch the needed IDs seq that did not match in the
	#database, and we store them in %IDs_out
	open(TMP, "Temporary_files/Notmatched") || die "Unable to open and create the tamporary file Notmatched needed\n";
	while(<TMP>) { chomp($_); $IDs_out{$_} = ""; }
	close(TMP);
	unlink("Temporary_files/Notmatched");
	#Then, we read all IDs from the treated sample, and keep only the
	#IDs_out, deleteing the others
	foreach $key (keys(%sample))
	{
		if(exists($IDs_out{$key})) { next; }
		else { delete($sample{$key}); delete($abundance{$key}); }
	}
	#We then return to the main program the needed references of hashes
	return(\%sample, \%abundance);
}
#End of function
#############################################################################

#############################################################################
# FUNCTION:		ClusteringAgainstDatabase
# USAGE:			ClusteringAgainstDatabase(ARGUMENTS)
# ARGUMENTS:	-$_[0]: the Input_args from user $choice_S,
# -$_[1]: the hash with otus from the DB,
# -$_[2]: the organized data (parallelized) in a hash table,
# -$_[3]: the reference of the %sample hash,
#	-$_[4]: the reference of the %abundance hash.
# DETAILS:		This function is defined to realize the clustering of sequences
# given by the user as input against the database chosen by the user. Several
# elements are needed as input, but the program give back no output.
#
# RELEASED: 10/10/2018
#############################################################################
sub ClusteringAgainstDatabase
{
	#Defined variables
	my %Input_args = %{$_[0]};
	my %otu = %{$_[1]};
	my %parallel_data = %{$_[2]};
	my %sample = %{$_[3]};
	my %gaps_db = %{$_[4]};

	my @keyotu = ();
	my $key = "";
	my $ref = "";
	my $output_header;
	my $nb_cores = 0;
	my $pid = 0;
	my @childs = ();
	my @IDs = ();
	my $ID = "";
	my $len_seq = 0;
	my $threshold = 0;
	my $myotu = "";
	my $count = 0;

	#Here, we launch the creation of the global output file and the Notmatched
	#temporary file to store out sequences
	open($output_header,">".$Input_args{'-o'}."Raw_clust_results.clust") || die "Unable to open and create the output file Raw_clust_results.clust needed\n";
	open(TMP, ">Temporary_files/Notmatched") || die "Unable to open and create the tamporary file Notmatched needed\n";
	#We then sort the DB sequences by OTU number
	@keyotu = sort(keys(%otu));
	#Here, using the parallel information, we treat each sequence
	LOOP01:foreach $key (keys(%parallel_data))
	{
		#We manage the limit of $cpu given by the user to treat data
		if($nb_cores >= $Input_args{"-cp"}) { $pid = wait(); $nb_cores --; }
		#We define a new process
		$pid = fork();
		#If the process is parent
		if($pid) { push(@childs, $pid); $nb_cores++; }
		#If the process is a child
		elsif ($pid == 0)
		{
			#We catch only needed data from %parallel_data hash
			@IDs = @{$parallel_data{$key}};
			#We take each analyzed sequence found in @IDs (from %parallel_data)
			LOOP02:foreach $ID (@IDs)
			{
				#We then compare the sample sequence with all sequences from the database
				LOOP03:foreach $ref (@keyotu)
				{
					#Here, we need to launch the Needleman-Wunsch alignment of the
					#user chose this solution
					if(uc($Input_args{"-cl"}) eq "NEEDLE")
					{
						($sample{$ID}, $otu{$ref}) = PerlLib::Alignment::NeedlemanWunschAlignment($sample{$ID}, $otu{$ref});
						$gaps_db{$ref} = $otu{$ref} =~ tr/\-/\-/;
					}

					#First we define the maximum number of differences tolerated and the length of the analyzed read
					if($myotu eq "") { ($len_seq, $threshold) = PerlLib::Clustering::DefineThresholdClust($gaps_db{$ref}, $sample{$ID}, $Input_args{"-th"}); }
					#Then, using the C function chosen based on the user choices, we launch the clustering itself.
					if(exists($Input_args{"-h"})) { $count = \PerlLib::CFunctions::c_compare_hm(uc($sample{$ID}),uc($otu{$ref}), $threshold); }
					elsif(exists($Input_args{"-c"})) { $count = \PerlLib::CFunctions::c_compare_len(uc($sample{$ID}),uc($otu{$ref}), $threshold); }
					elsif(exists($Input_args{"-hc"})) { $count = \PerlLib::CFunctions::c_compare_hm_len(uc($sample{$ID}),uc($otu{$ref}), $threshold); }
					else { $count = \PerlLib::CFunctions::c_compare(uc($sample{$ID}),uc($otu{$ref}), $threshold); }
					#To fast the analysis, and keep only the best result, we redefine the threshold if a hit is found
					if(${$count} < $threshold) { $threshold = ${$count}; $myotu = $ref; }
					#To fast the analysis, if a read have a 100% similarity, we stop the analysis for this result
					if($threshold == 0)
					{
						$myotu = PerlLib::Clustering::CompileClusteringResult($output_header, $ID, $myotu, $gaps_db{$myotu}, $sample{$ID}, $Input_args{"-th"}, $threshold);
						next LOOP02;
					}
				}
				#If we found nothing higher than the chosen threshold of the user
				if($myotu eq "") { printf(TMP "$ID\n"); next LOOP02; }
				#If we found something, but not with 100% of identity but higher than the defined threshold
				else { $myotu = PerlLib::Clustering::CompileClusteringResult($output_header, $ID, $myotu, $gaps_db{$myotu}, $sample{$ID}, $Input_args{"-th"}, $threshold); }
			}
			exit(0);
		}
		#If we can't fork
		else { die "coudn't fork !\n"; }
	}
	close(TMP);
	close($output_header);
	#We wait for the complete treatment of all children
	foreach (@childs) { waitpid($_, 0);	}
}
#End of function
#############################################################################

#############################################################################
# FUNCTION:		ClusteringDeNovo
# USAGE:			ClusteringDeNovo(ARGUMENTS)
# ARGUMENTS:	-$_[0]: the Input_args from user choices,
# -$_[1]: the reference of the %sample hash,
#	-$_[2]: the reference of the %abundance hash.
# DETAILS:		This function is defined to realize the clustering of sequences
# given by the user as input by a de novo approach. Several elements are needed
# as input, but the program give back no output.
#
# RELEASED: 10/10/2018
#############################################################################
sub ClusteringDeNovo
{
	#Defined variables
	my %Input_args = %{$_[0]};
	my %sample = %{$_[1]};
	my %abundance = %{$_[2]};
	my %otu = ();
	my $ref = "";
	my $ID = "";
	my %gaps_db = ();
	my @IDs = ();
	my $output_header;
	my %out = ();
	my $len_seq = 0;
	my $threshold = 0;
	my $count = 0;

	#Here, as the user chose the INFERNAL alignment option, we define for each
	#sequence not affiliated to the DB, their number of gaps, to fast the analysis
	if(uc($Input_args{"-cl"}) eq "INFERN")
	{
		$ref = PerlLib::Alignment::DefineGaps($_[1]);
		%gaps_db = %{$ref};
	}
	#Then, we sort, based on their abundance in the dataset, the sequence IDs
	@IDs = keys(%sample);
	@IDs = sort { $abundance{$b} <=> $abundance{$a} || $a cmp $b } @IDs;

	open($output_header,">>".$Input_args{"o"}."Raw_clust_results.clust") || die "Unable to open and create the output file Raw_clust_results.clust needed\n";
	#For each sequence that needs treatment
	LOOP01:foreach $ID (@IDs)
	{
		#If we don't have yet a seed OTU (first sequence of a de novo clustering)
		if (keys(%otu) == 0)
		{
			#We create it using the first sequence from the @IDs data
			$otu{$ID} = $sample{$ID};
			PerlLib::Clustering::CompileClusteringResult($output_header, $ID, $ID, $gaps_db{$ID}, $sample{$ID}, $Input_args{"-th"}, 0);
			next LOOP01;
		}
		#We then compare the sequence with all seed sequnces from the %otu enriched during the analysis
		LOOP02:foreach $ref (keys %otu)
		{
			#Here, we need to launch the Needleman-Wunsch alignment of the user chose this solution
			if(uc($Input_args{"-cl"}) eq "NEEDLE")
			{
				($sample{$ID}, $otu{$ref}) = PerlLib::Alignment::NeedlemanWunschAlignment($sample{$ID}, $otu{$ref});
				$gaps_db{$ref} = $otu{$ref} =~ tr/\-/\-/;
			}

			#First we define the maximum number of differences tolerated and the length of the analyzed read
			($len_seq, $threshold) = PerlLib::Clustering::DefineThresholdClust($gaps_db{$ref}, $sample{$ID}, $Input_args{"-th"});
			#Then, using the C function chosen based on the user choices, we launch the clustering itself.
			if(exists($Input_args{"-h"})) { $count = \PerlLib::CFunctions::c_compare_hm(uc($sample{$ID}),uc($otu{$ref}), $threshold); }
			elsif(exists($Input_args{"-c"})) { $count = \PerlLib::CFunctions::c_compare_len(uc($sample{$ID}),uc($otu{$ref}), $threshold); }
			elsif(exists($Input_args{"-hc"})) { $count = \PerlLib::CFunctions::c_compare_hm_len(uc($sample{$ID}),uc($otu{$ref}), $threshold); }
			else { $count = \PerlLib::CFunctions::c_compare(uc($sample{$ID}),uc($otu{$ref}), $threshold); }
			#If we found a result higher than the defined threshold, we print the result and start to analyze a new sequence
			if(${$count} < $threshold)
			{
				PerlLib::Clustering::CompileClusteringResult($output_header, $ID, $ref, $gaps_db{$ref}, $sample{$ID}, $Input_args{"-th"}, ${$count});
				next LOOP01;
			}
		}
		#If, at the end of the comparison with all seeds, the read did not match anything, it becomes a new seed
		$otu{$ID} = $sample{$ID};
		PerlLib::Clustering::CompileClusteringResult($output_header, $ID, $ID, $gaps_db{$ID}, $sample{$ID}, $Input_args{"-th"}, 0);
		next LOOP01;
	}
	close($output_header);
}
#End of function
#############################################################################
1;
