###############################################################################
#POD Part
###############################################################################

=head1 NAME

package I<EnrichDB> - library used to realize the treatment and enrichment of
the used database with new reads and data.

=head1 LIBRARY ORGANIZATION AND FUNCTIONS

	use EnrichDB;
	ManageEnrichDB();
  ExtractNewSeeds();
  ExtractNumsDB();
  EnrichDBFiles();

=head1 DESCRIPTION

This module was defined and used to extract new seeds defined after comparison
with the chosen database, define their new names and integrate them into the
different files of the database.
The documentation was realized with B<Pod>.

=head2 FUNCTIONS

=over 4

=item B<ManageEnrichDB()>

This function is dedicated to manage the extraction and enrichment of the chosen
database with new seeds and new data from the analysis. It takes as input
several elements to launch the different steps of treatment.

=item B<ExtractNewSeeds()>

This function is dedicated to the extraction of new sequences from the treated
file. More precisely, it extracts the IDs and corresponding reads to add
efficiently new seeds to the database.

=item B<ExtractNumsDB()>

This function is dedicated to the extraction of data from database files. More
precisely, it extracts the number of samples and OTUs to add efficiently new
seeds.

=item B<EnrichDBFiles()>

This function is dedicated to the addition of sequences to the dbreads file, the
data into the database file and the creation of a new global alignment, stored
in the dedicated folder.

=back

=head1 AUTHORS

Sebastien TERRAT <sebastien.terrat@inra.fr>
Corentin Journay

=head1 TEAM AND NOTES

UMR 1347 INRA/uB/AgroSup - BioCom Team
RELEASED:  30/10/2018

=cut

###############################################################################
#FUNCTIONS AND PROGRAMS
###############################################################################

# Name of the package defined for the main program
package PerlLib::EnrichDB;
use strict;
use warnings;

###############################################################################
# FUNCTION:		ManageEnrichDB
# USAGE:			ManageEnrichDB(ARGUMENTS)
# ARGUMENTS:	- The Fasta file given as input by the user.
#							- The database folder given as input by the user.
#             - The number of cpus chosen by the user.
#             - The folder where results will be stored.
# DETAILS:		This function is dedicated to manage the extraction and enrichment
#             of the chosen database with new seeds and new data from the
#             analysis.
# RELEASED: 	10/10/2018
###############################################################################
sub ManageEnrichDB
{
  #Defined variables
  my $sample = $_[0];
  my $db_folder = $_[1];
  my $nb_cpus = $_[2];
  my $output_folder = $_[3];
  my @tmp = ();
  my @tmp2 = ();
  my @tmp3 = ();
  my %seqs = ();
  my %OTUs = ();
  my %samples = ();
  my $key = "";
  my $value_otu = 0;
  my $value_sample = 0;
  my $filename = "";
  my $database = "";
  my $ref_array = "";

  #First, we catch the reads considered as seeds for new OTUs after the de novo
  #approach done by the main program
  $ref_array = PerlLib::EnrichDB::ExtractNewSeeds($sample, $output_folder);
  #Then, we found and defined the OTU number and sample number that will be used to
  #enrich the database afterwards.
  ($value_otu, $value_sample) = PerlLib::EnrichDB::ExtractNumsDB($db_folder);
  #Finally, we add, append or create needed data to enrich the three files of the database
  PerlLib::EnrichDB::EnrichDBFiles($ref_array, $db_folder, $value_otu, $value_sample, $nb_cpus);
}
#End of function
###############################################################################

###############################################################################
# FUNCTION:		ExtractNewSeeds
# USAGE:			ExtractNewSeeds(ARGUMENTS)
# ARGUMENTS:	- The Fasta file given as input by the user ($_[0]).
#							- The database folder given as input by the user ($_[1]).
# DETAILS:		This function is dedicated to the extraction of data from database
# files. More  precisely, it extracts the number of samples and OTUs to add
# efficiently new seeds.
# RELEASED: 	10/10/2018
###############################################################################
sub ExtractNewSeeds
{
  #Defined variables
  my $filename = $_[0];
  my $output_folder = $_[1];
  my @tmp = ();
  my %seeds = ();

  #Reading the Raw result file after clustering (against the database and the de novo steps)
  open(F_RAW, $output_folder."Raw_clust_results.clust") || die "Unable to open the output file Raw_clust_results.clust needed\n";
  #First, we manage the file of raw results to obtain data
  while(<F_RAW>)
  {
    #We defined which sequences we need to catch and add to the database
    if($_ =~ m/\tOTU/) { next; }
    else { chomp($_); @tmp = split("\t", $_); $seeds{$tmp[2]} = ""; }
  }
  #We then close the file
  close(F_RAW);
  #We then catch the correspond seeds from the input initial file
  open(F_SEQ,"$filename") || die "Unable to open the $filename file needed\n";
  while(<F_SEQ>)
  {
    $_ =~ m/^>(.*)\n/;
    if(exists($seeds{$1})) { $seeds{$1} = <F_SEQ>; }
  }
  close(F_SEQ);
  #We finally return the reads considered as seeds associated to their ID
  return(\%seeds);
}
#End of function
###############################################################################

###############################################################################
# FUNCTION:		ExtractNumsDB
# USAGE:			ExtractNumsDB(ARGUMENTS)
# ARGUMENTS:	- The Fasta file given as input by the user ($_[0]).
#							- The database folder given as input by the user ($_[1]).
# DETAILS:		This function is dedicated to the extraction of data from two
# files (DB and given as input).
# RELEASED: 	10/10/2018
###############################################################################
sub ExtractNumsDB
{
  #Defined variables
  my $db_folder = $_[0];
  my $filename = "";
  my $database = "";
  my $key = "";
  my $value_otu = 0;
  my $value_sample = 0;
  my @tmp = ();
  my @tmp2 = ();
  my @tmp3 = ();
  my %OTUs = ();
  my %samples = ();

  #First, we catch and store the name of the database file.
  opendir (DIR, "$db_folder/") || die "can't open the directory $db_folder: $!";
  LOOP01:while (defined($filename = readdir(DIR))) { if ($filename =~ m/^database_/) { $database = $filename; last LOOP01; } }
  closedir(DIR);
  #We opened it to found the biggest OTU number and also the biggest sample number into the current database
  open(DB, "$db_folder/$database") || die "can't open the file $database in the directory $db_folder\n";
  while(<DB>)
  {
    #To do this, we search into the data to find these information
    chomp($_);
    @tmp = split("\t", $_);
    $OTUs{$tmp[0]} = "";
    @tmp2 = split(/\|/, $tmp[3]);
    foreach $key (@tmp2) { @tmp3 = split("_", $key); $samples{$tmp3[1]} = ""; }
  }
  close(DB);

  #Then, using organized data, we defined the highest values for OTU and sample
  #numbers, to return them to the main function.
  @tmp = sort{$a cmp $b} keys(%OTUs);
  $tmp[$#tmp] =~ /OTU(\d*)/;
  $value_otu = sprintf("%09d", $1 + 1);
  @tmp2 = sort{$a <=> $b} keys(%samples);
  $value_sample = sprintf("%09d", $tmp2[$#tmp2] + 1);
  #We then return found results to the main function
  return($value_otu, $value_sample);
}
#End of function
###############################################################################

###############################################################################
# FUNCTION:		EnrichDBFiles
# USAGE:			EnrichDBFiles(ARGUMENTS)
# ARGUMENTS:	- The seed sequences that will be added ($_[0]).
#							- The database folder given as input by the user ($_[1]).
#             - The number of already present OTUs in the database ($_[2]).
#             - The number of already present samples in the database ($_[3]).
#             - The number of cpus chosen by the user ($_[4]).
# DETAILS:		This function is dedicated to the addition of sequences to the
# dbreads file, the data into the database file and the creation of a new global
# alignment, stored in the dedicated folder.
# RELEASED: 	10/10/2018
###############################################################################
sub EnrichDBFiles
{
  #Defined variables
  my %seeds = %{$_[0]};
  my $db_folder = $_[1];
  my $value_otu = $_[2];
  my $value_sample = $_[3];
  my $nb_cpus = $_[4];
  my $ID = "";
  my $filename = "";
  my $dbreads = "";
  my $dbalign = "";
  my $database = "";
  my $tmp_file = "";

  #First, we catch and store the name of the three database files.
  opendir (DIR, "$db_folder/") || die "can't open the directory $db_folder: $!";
  LOOP01:while (defined($filename = readdir(DIR)))
  {
    if($filename =~ m/^database_/) { $database = $filename; }
    elsif($filename =~ m/^dbreads/) { $dbreads = $filename; }
    elsif($filename =~ m/^dbalign/) { $dbalign = $filename; }
  }
  closedir(DIR);
  #Then, we appended the two files dbreads and database, and we also created a TMP file to realize
  #the global alignment of new seeds to merge both alignments into a new complete file.
  open(F_DBREADS, ">>$db_folder/$dbreads") || die "Can't append the $db_folder/$dbreads file\n";
  open(F_DB, ">>$db_folder/$database") || die "Can't append the $db_folder/$database file\n";
  open(TMP, ">Temporary_files/TMP_seeds") || die "Can't create the Temporary_files/TMP_seeds file\n";
  #For each read added to the database
  foreach $ID (keys(%seeds))
  {
    #We searched for its abundance
    $ID =~ m/_(\d*)_/;
    #We defined the FASTA files (dbreads and the temporary file)
    printf(F_DBREADS ">OTU".$value_otu."\n".$seeds{$ID});
    printf(TMP ">OTU".$value_otu."\n".$seeds{$ID});
    #We also append the database file
    print(F_DB "OTU".$value_otu."\t".$1."\t1.00_Unknown|1.00_Unknown|1.00_Unknown|1.00_Unknown|1.00_Unknown\t".$1."_$value_sample\n");
    $value_otu++;
  }
  close(F_DBREADS);
  close(F_DB);
  close(TMP);
  #Here, we launched the global alignment of the seeds that will be added to the dbalign file
  $tmp_file = PerlLib::Alignment::LaunchInfernalAlign("Temporary_files/TMP_seeds", $db_folder, $nb_cpus, "Temporary_files/TMP_seeds");
  unlink("Temporary_files/TMP_seeds");
  #Then, we defined a new global alignment file to replace the old one
  $tmp_file = PerlLib::Alignment::MergeAlignments($db_folder, $tmp_file, "Temporary_files/$dbalign");
  unlink("$db_folder/$dbalign");
  unlink("Temporary_files/Align_TMP_seeds");
  #Finally, we opened the temporay file, to add to the new one data
  open(F_ALIGN, ">$db_folder/$dbalign") || die "Can't append the $db_folder/$dbalign file\n";
  open(F_ALIGN_TMP, "$tmp_file") || die "Can't read the $tmp_file file\n";
  while(<F_ALIGN_TMP>) { printf(F_ALIGN $_); }
  close(F_ALIGN_TMP);
  close(F_ALIGN);
  #We finally delete not needed files
  unlink($tmp_file);
}
1;
