###############################################################################
#POD Part
###############################################################################

=head1 NAME

package I<Formatting> - library used to realize the treatment and enrichment of
the used database with new reads and data.

=head1 LIBRARY ORGANIZATION AND FUNCTIONS

	use Formatting;
	FormattingResults();

=head1 DESCRIPTION

This module was defined to create all summary and result files given as output
for the user. It treats the Raw file and give back counts and useful data for
easy analysis.
The documentation was realized with B<Pod>.

=head2 FUNCTIONS

=over 4

=item B<FormattingResults()>

This function is dedicated to the definition of various summary results (such as
the number of reads associated to OTUs from the database or using a de novo
approach). A summary file giving all these information is produced, and also a
matrix file easily used in R, for example. Finally, a file linking the taxonomy
of each OTU with compared reads is also given for the user.

=back

=head1 AUTHORS

Sebastien TERRAT <sebastien.terrat@inra.fr>
Corentin Journay

=head1 TEAM AND NOTES

UMR 1347 INRA/uB/AgroSup - BioCom Team
RELEASED:  30/10/2018

=cut

###############################################################################
#FUNCTIONS AND PROGRAMS
###############################################################################

# Name of the package defined for the main program
package PerlLib::Formatting;
use strict;
use warnings;


###############################################################################
# FUNCTION:		FormattingResults
# USAGE:		  FormattingResults(ARGUMENTS)
# ARGUMENTS:	- $_[0]: filename given as input by the user
#             - $_[1]: the chosen database used
#             - $_[2]: the output folder used to store results
# DETAILS:	This function is dedicated to the definition of various summary
#           results (such as the number of reads associated to OTUs from the
#           database or using a de novo approach). A summary file giving all
#           these information is produced, and also a matrix file easily used
#           in R, for example. Finally, a file linking the taxonomy of each OTU
#           with compared reads is also given for the user.
# RELEASED: 10/10/2018
###############################################################################
sub FormattingResults
{
  #Defined variables
  my $sample = $_[0];
  my $db_folder = $_[1];
  my $output_folder = $_[2];
  my @tmp = ();
  my @tmp2 = ();
  my @keys = ();
  my %OTUs = ();
  my %OUTs = ();
  my ($nbseqtot, $nbseqotus, $nbseqouts) = (0,0,0);
  my ($nbotustot, $nbotus, $nbouts) = (0,0,0);
  my $key = "";
  my $filename = "";
  my $database = "";
  my $i = sprintf("%09d", 1);

  #Reading the Raw result file after clustering (against the database and the de novo steps)
  open(F_RAW,"$output_folder"."Raw_clust_results.clust") || die "Unable to open the output file Raw_clust_results.clust needed\n";
  #First, we manage the file of raw results to obtain data
  while(<F_RAW>)
  {
    #We cleaned and catched the data
    chomp($_);
    @tmp = split("\t", $_);
    @tmp2 = split("_", $tmp[0]);
    #We count the total number of reads
    $nbseqtot += $tmp2[1];
    #We defined the number of OTUs and OUTs obtained and their corresponding
    #number of reads
    if($tmp[2] =~ m/OTU/) { $OTUs{$tmp[2]} += $tmp2[1]; $nbseqotus += $tmp2[1]; }
    else { $OUTs{$tmp[2]} += $tmp2[1]; $nbseqouts += $tmp2[1]; }
  }
  close(F_RAW);
  #We first count the number of reads and OTUs in both cases
  $nbotus = keys(%OTUs);
  $nbouts = keys(%OUTs);
  $nbotustot = $nbotus + $nbouts;
  #We then print this data into the summary file containing summarized results
  open(F_SUM, ">".$output_folder."Summary_results.txt") || die "Unable to create the output summary file Summary_results.txt needed\n";
  printf(F_SUM "Total number of reads:\t$nbseqtot\nNumber of reads associated to OTUs from the database:\t$nbseqotus\n");
  printf(F_SUM "Number of reads associated to OTUs de novo\t$nbseqouts\nTotal number of OTUs (database and de novo):\t$nbotustot\n");
  printf(F_SUM "Number of OTUs associated to OTUs from the database:\t$nbotus\nNumber of OTUs defined de novo:\t$nbouts\n");
  close(F_SUM);
  #Then, we organized by order the OTUs to print needed data for the user
  @keys = sort{$a cmp $b} keys(%OTUs);
  @tmp = sort{$OUTs{$a} <=> $OUTs{$b} or $a cmp $b} keys(%OUTs);
  #Based on the organized data, we printed the matrix containing data
  #from both clustering approaches, organized for the user, in a dedicated file
  open(F_MAT, ">".$output_folder."Matrix_data.txt") || die "Unable to create the output file Matrix_data.txt needed\n";
  printf(F_MAT "\t".join("\t", @keys));
  foreach $key (@tmp) { printf(F_MAT "\tOUT$i"); $i++; }
  printf(F_MAT "\n$sample");
  foreach $key (@keys) { printf(F_MAT "\t$OTUs{$key}"); }
  foreach $key (@tmp) { printf(F_MAT "\t$OUTs{$key}"); }
  printf(F_MAT "\n");
  close(F_MAT);
  #Then, we catch and store the name of the database file.
  opendir (DIR, "$db_folder/") || die "can't open the directory $db_folder: $!";
  LOOP01:while (defined($filename = readdir(DIR))) { if ($filename =~ m/^database_/) { $database = $filename; last LOOP01; } }
  closedir(DIR);
  #We opened it to give back to the user needed information
  open(DB, "$db_folder/$database") || die "can't open the file $database in the directory $db_folder\n";
  open(F_DB, ">".$output_folder."Data_from_db_$database.txt") || die "Unable to create the output file Data_from_db_$database.txt needed\n";
  #We then print needed data for the user to link information from the database to the analyzed sample.
  printf(F_DB "OTU\tNumber of reads from your sample\tNumber of reads for this OTU in the database\tDefined taxonomy\tSample representativity\n");
  #Finally, we print reorganized data from the database to link input reads with potentil taxonomies
  #of OTUs from the database.
  while(<DB>)
  {
    chomp($_);
    @tmp = split("\t", $_);
    #We check if the OTU exists
    if(exists($OTUs{$tmp[0]}))
    {
      printf(F_DB $tmp[0]."\t".$OTUs{$tmp[0]}."\t".$tmp[1]."\t".$tmp[2]."\t".$tmp[3]."\n");
    }
  }
  close(DB);
  close(F_DB);
}
#End of function
###############################################################################
1;
