###############################################################################
#POD Part
###############################################################################

=head1 NAME

package I<TreatDataForClust> - library used to realize previous and .

=head1 LIBRARY ORGANIZATION AND FUNCTIONS

	use TreatDataForClust;
	ManagingInputData();
	ConcatDBTreatedFile();
	ExtractDataFromFiles();
	DefineParallelizedHashes();

=head1 DESCRIPTION

This module was defined and used to realize, compile and clean results from
the clustering step. This module contains functions to realize the clustering,
both against the database and de novo. But also functions to compile results and
cleaning data.
The documentation was realized with B<Pod>.

=head2 FUNCTIONS

=over 4

=item B<ManagingInputData()>

This function is dedicated to merge both files (DB and input), launch if needed,
the multiple alignment, and extract needed data for further analysis. Give back
as output references to hashes for further analyses. It needs only user choices
as input.

=item B<ConcatDBTreatedFile()>

This function is dedicated to the merging of the database file and the FASTA
file that will be treated before the global alignment using the INFERNAL
program of alignment. As output, the function will return the newly created file
already aligned.

=item B<ExtractDataFromFiles()>

This function is dedicated to the extraction of data from two files (DB and
given as input). It returns as output three hash references (otu, sample and
abundance).

=item B<DefineParallelizedHashes()>

This function is dedicated to the defintion of a hash containing splitted data
to treat in parallel the clustering of sequences against the database used and
defined by the user. As output, the reference to this hash table is returned to
the main program.

=back

=head1 AUTHORS

Sebastien TERRAT <sebastien.terrat@inra.fr>
Corentin Journay

=head1 TEAM AND NOTES

UMR 1347 INRA/uB/AgroSup - BioCom Team
RELEASED:  30/10/2018

=cut

###############################################################################
#FUNCTIONS AND PROGRAMS
###############################################################################

# Name of the package defined for the main program
package PerlLib::TreatDataForClust;
use strict;
use warnings;

###############################################################################
# FUNCTION:		ManagingInputData
# USAGE:		  ManagingInputData(ARGUMENTS)
# ARGUMENTS:	- $_[0]: The reference of the hash table storing user choices
# DETAILS:		This function is dedicated to merge both files (DB and input),
# launch if needed, the multiple alignment, and extract needed data for further
# analysis.
# RELEASED: 10/10/2018
###############################################################################
sub ManagingInputData
{
  #Defined variables
  my %Input_args = %{$_[0]};
  my $tmp_file = "";
  my ($ref1, $ref2, $ref3, $ref4) = ("","","","");
  my $time = time;

  #Here, as the user chose the INFERNAL alignment option, we will realize the alignment
  #and treatment of the given file and the database
  if(uc($Input_args{"-cl"}) eq "INFERN")
  {
    print ("STARTING GLOBAL ALIGNMENT OF THE DATABASE AND THE SEQUENCES\n");
  	#First, we concatenate both sequence files (DB and input)
  	$tmp_file = PerlLib::TreatDataForClust::PrepareTreatedFile($Input_args{"-sq"});
  	#We launch the realization of the global alignment for the input reads
  	$tmp_file = PerlLib::Alignment::LaunchInfernalAlign($tmp_file, $Input_args{"-db"}, $Input_args{"-cp"}, "");
    #Then, we merge both alignment files (from the database and the input reads)
    $tmp_file = PerlLib::Alignment::MergeAlignments($Input_args{"-db"}, $tmp_file, "Temporary_files/Align_file");
  	#We then sort and organize data from the DB and the treated file
  	($ref1, $ref2, $ref3) = PerlLib::Alignment::ExtractDataFromAlignment($tmp_file);
  	print ("END OF GLOBAL ALIGNMENT AND EXTRACTING DATA\n");
  	print (time-$time." Seconds of treatment"."\n");
  	#Here, we define for each representative sequence of the DB the number of gaps, to fast the analysis
  	$ref4 = PerlLib::Alignment::DefineGaps($ref1);
  }
  #Here, as the user chose the NEEDLEMAN-WUNSCH alignment option, we just have to catch the sequences from
  #the given file and the database file before launching the clustering itself
  elsif(uc($Input_args{"-cl"}) eq "NEEDLE")
  {
  	#First, as we need only the sequences, we just read and store the data
  	($ref1, $ref2, $ref3) = PerlLib::TreatDataForClust::ExtractDataFromFiles($Input_args{"-sq"}, $Input_args{"-db"});
  }
  return($ref1, $ref2, $ref3, $ref4);
}

###############################################################################
# FUNCTION:		PrepareTreatedFile
# USAGE:		  PrepareTreatedFile(ARGUMENTS)
# ARGUMENTS:	- $_[0]: The Fasta file given as input by the user.
# DETAILS:		  This function is dedicated to the addition of a prefix to
#     sequences given as input by the user to identify them into the global
#     alignment using the INFERNAL program of alignment. As output, the
#			function will return the newly created file aligned.
# RELEASED: 10/10/2018
###############################################################################
sub PrepareTreatedFile
{
	#Defined variables
	my $seq_file = $_[0];

	#We first create the temporary file containing the reads that will be modified
	open(F_TMP, ">Temporary_files/TMP_seq_file") || die "Can't create the TMP_seq_file in the Temporary_files directory, please check.\n";
  #Then, we add to the Temporary file created the sequences from the treated file
	open(F_SQ, $seq_file) || die "Can't open the $seq_file given as input by the user, please check.\n";
	while(<F_SQ>)
	{
		if($_ =~ s/^>/>SQ_/) { printf(F_TMP $_); }
		else { printf(F_TMP $_); }
	}
	close(F_SQ);
	close(F_TMP);
	#We finally return the TMP_seq_file name for further use
	return("Temporary_files/TMP_seq_file");
}
#End of function
###############################################################################

###############################################################################
# FUNCTION:		ExtractDataFromFiles
# USAGE:			ExtractDataFromFiles()
# ARGUMENTS:	- The Fasta file given as input by the user ($_[0]).
#							- The database folder given as input by the user ($_[1]).
# DETAILS:		This function is dedicated to the extraction of data from two
# files (DB and given as input).
# RELEASED: 	10/10/2018
###############################################################################
sub ExtractDataFromFiles
{
	#Defined variables
	my $seq_file = $_[0];
	my $DB_folder = $_[1];
	my %otu = ();
	my %sample = ();
	my %abundance = ();
	my @tmp = ();
	my $ID = "";
	my $filename = "";
	my $dbreads = "";

	#First, we catch and store the name of the dbreads file.
	opendir (DIR, "$DB_folder/") || die "can't open the directory $DB_folder: $!";
	LOOP01:while (defined($filename = readdir(DIR))) { if ($filename =~ m/^dbreads_/) { $dbreads = $filename; last LOOP01; } }
	closedir(DIR);
	#We then extract from the database file and store it in %otu
	open(F_DB, $DB_folder.$filename) || die "Can't open the $filename from the $DB_folder directory, please check.\n";
	while(<F_DB>)
	{
		chomp($_);
		if($_ =~ s/^>//) { $ID = $_; }
		else { $otu{$ID} = $_; $ID = ""; }
	}
	close(F_DB);
	#We then extract from the treated file and store it in %sample and %abundance
	open(F_SQ, $seq_file) || die "Can't open the $seq_file given as input by the user, please check.\n";
	while(<F_SQ>)
	{
		chomp($_);
		if($_ =~ s/>//)
		{
			$ID = $_;
			@tmp = split("_", $ID);
			$abundance{$ID} = $tmp[1];
		}
		else { $sample{$ID} = $_; $ID = ""; }
	}
	close(F_SQ);
	#We finally return the three filled hashes
	return(\%otu,\%sample,\%abundance);
}
#End of function
###############################################################################

###############################################################################
# FUNCTION:		DefineParallelizedHashes
# USAGE:			DefineParallelizedHashes()
# ARGUMENTS:	-$_[0]: the reference of th hash table %abundance,
#             -$_[1]: the number of cpus defined by the user (or by default).
# DETAILS:		This function is dedicated to the defintion of a hash containing
# splitted data to treat in parallel the clustering of sequences against the
# database used and defined by the user. As output, the reference to this hash
# table is returned to the main program.
# RELEASED: 	10/10/2018
###############################################################################
sub DefineParallelizedHashes
{
  #Defined variables
  my $ref_abundance = $_[0];
  my $nb_cpus = $_[1];
  my @sample_IDs = ();
  my %parallel_data = ();
  my ($i,$j) = (0,0);

  #We sort the sample sequence IDs
  @sample_IDs = sort(keys(%{$ref_abundance}));
  #Here, we split the data based on the number of cpu, to treat efficiently the sample
  for($i=1;$i<=$nb_cpus;$i++)
  {
  	$parallel_data{$i} = [@sample_IDs[$j..($#sample_IDs/$nb_cpus)*$i]];
  	$j = ($#sample_IDs/$nb_cpus)*$i + 1;
  }
  #Based on this splitting, we return the hash table that will be treated during
  #the clustering
  return(\%parallel_data);
}
1;
