###############################################################################
#POD Part
###############################################################################

=head1 NAME

package I<TreatDataForDb> - library used to treat file(s) given as input by the
user to define a new database containing sequences and their corresponding data
to efficiently compare new samples to this database.

=head1 LIBRARY ORGANIZATION AND FUNCTIONS

	use TreatDataForDb;
	ExtractClustDerepData();
	InvDerepClustData();
	InvDerepClustData2(); ??
	ExtractAndStoreSeedSeq();
	ExtractTaxonomicData();
	DefineDatabaseFile();

=head1 DESCRIPTION

This module was defined to extract from each file given as input by the user and
checked by the program needed information, to define both needed files for the
database definition: the seed sequences, and the linked information (number of
reads, samples, taxonomic information, etc.).
The documentation was realized with B<Pod>.

=head2 FUNCTIONS

=over 4

=item B<ExtractClustDerepData()>

This function was developed to read and store data from two input files: the
clustering file and the dereplication file. As output, this function returns
four references to four hash tables needed for subsequent analyses:
%corre I<$key>: line of clustering composition, $value: ID seed,
%clust: I<$key>: seed ID, $value: number of reads in the OTU,
%IDs: I<$key>: seed ID, $value: cluster ID,
%derep_info: I<$key>: sequence ID chosen after dereplication,
$value: corresponding unique reads.

=item B<InvDerepClustData()>

Use data from the Dereplication file to redefine clusters. This function was
developed to analyze all clusters details and replace all ID of dereplicated
reads by their corresponding unique reads using the %derep_info hash. The
modifications are stored in %corre and %clust. These three variables are finally
given as reference outputs to the main program.

=item B<DefineSamplingDistribution()>

Define the composition fo each OTU. This function was developed to treat the
information from each cluster using the %corre hash table as input, and give
back a line of description (nbreads_sampleID|nbreads_sample_ID...) to describe
the composition of each OTU in the database. It takes as input a reference to
%corre hash ($key: line of clustering composition, $value: ID seed).

=item B<ExtractAndStoreSeedSeq()>

Extract from the Sequence file each seed sequence. This function is dedicated
to the search and storage of the seed sequences to create the complete database.
Based on several inputs (sequence filename, %IDs hash, folder name of the
database and user choices (@choices), this function seek all needed reads, check
if nothing is lost, and create the dbreads file. It returns the reference to the
%IDs hash table modified and completed, and create the file containing sequences
chosen for the database (seed sequences for each OTU).

=item B<ExtractTaxonomicData()>

Extract and treat from the Taxonomic file needed data. This function was
developed to find, format and extract all taxonomic details for each OTU.
This function needs only the taxonomic filename as input, and gives back as
output a hash table with OTU ID as key and as value the details of taxonomy,
associating tha $key: OTU ID, and the $value: taxonomic details.

=item B<DefineDatabaseFile()>

Define the file containing all data needed the database creation. This program
takes as input six arguments and more particularly the:
%clust: I<$key>: seed ID, $value: number of reads in the OTU,
%glob_distrib: I<$key>: seed ID for the cluster, $value : line of info for db,
%IDs: I<$key>: seed ID, $value: cluster ID,
%Taxo_IDs: I<$key>: OTU ID, $value: taxonomic details.
Moreover, the function takes also the folder defined to store the database, and
the user choices stored in an array. As output, this function create efficiently
the database file giving details for each seed sequence.

=back

=head1 AUTHORS

Sebastien TERRAT <sebastien.terrat@inra.fr> and Corentin Journay

=head1 TEAM AND NOTES

UMR 1347 INRA/uB/AgroSup - BioCom Team
RELEASED:  30/10/2018

=cut

###############################################################################
#FUNCTIONS AND PROGRAMS
###############################################################################

# Name of the package defined for the main program
package PerlLib::TreatDataForDb;
use strict;
use warnings;

###############################################################################
#FUNCTION:	ExtractClustDerepData
#USAGE:			ExtractClustDerepData(ARGUMENTS)
#ARGUMENTS:	-$_[0]: reference to @choices (user input choices)
#						-$_[1]: reference to %Input_args hash (input files)
#DETAILS:	This function was developed to read and store data from two
#		input files: the clustering file and the dereplication file.
#		As output, this function returns three references to four hash
#		tables needed for subsequent analyses.
#RELEASED:	09/10/2017
###############################################################################
sub ExtractClustDerepData
{
	#Defined variables
	my @choices = @{$_[0]};
	my %Input_args = %{$_[1]};
	my @tmp = ();
	my @id = ();
	my %corre = (); #$key: line of clustering composition, $value: ID seed
	my %clust = (); #$key: seed ID, $value: number of reads in the OTU
	my %IDs = (); #$ey: seed ID, $value: cluster ID
	my %derep_info = (); #$key: sequence ID chosen after dereplication, $value: corresponding unique reads

	#For each line found in the clustering file
	open(CLUST,"$Input_args{-clust}")||die "==> Unable to open $Input_args{-clust} file\n.";
	LOOP01:while(<CLUST>)
	{
		if($_=~m/^\d/)
		{
			@tmp = split("\t",$_);
			#Here, we avoid the conservation of single-singletons OTUs
			if($choices[1] eq "NO" && $tmp[2] == 1) { next LOOP01; }
			#Then, we extract the ID of the seed
			$tmp[3] =~ m/\d*\|(.*?)\s/;
			#Using this seed ID, we fill the %corres, %IDs and %clust hashes
			$corre{$tmp[3]}=$1;
			$IDs{$1} = $tmp[0];
			$clust{$1} = $tmp[2];
		}
	}
	close(CLUST);

	#For each line found in the dereplication file
	open(DEREP,"$Input_args{-derep}")||die "==> Unable to open $Input_args{-derep} file\n.";
	#For each line found in the dereplication file
	LOOP02:while(<DEREP>)
	{
		$_ =~ s/\n//;
		@tmp = split("\t",$_);
		#We associated one unique sequence to all dereplicated ones
		$derep_info{$tmp[0]}=$tmp[1];
	}
	close(DEREP);
	#We finally return the filled arrays
	return(\%corre, \%clust, \%derep_info, \%IDs);
}
#End of function
###############################################################################

###############################################################################
#FUNCTION:	InvDerepClustData
#USAGE:	InvDerepClustData(@ref_arrays)
#ARGUMENTS:	-$_[1]: reference to %corre hash (seed ID linked to clustering)
#		-$_[2]: reference to %clust hash (seed ID linked to nb reads)
#		-$_[3]: reference to %derep_info hash (derep ID linked to unique ID)
#DETAILS:	This function was developed to analyze all clusters details and
#		replace all ID of dereplicated reads by their corresponding
#		unique reads using the %derep_info hash. The modifications are
#		stored in %corre and %clust. These three variables are finally
#		given as output to the main program.
#RELEASED: 09/10/2017
###############################################################################
sub InvDerepClustData
{
	#Defined variables
	my %corre = %{$_[0]}; #$key: line of clustering composition, $value: ID seed
	my %clust = %{$_[1]}; #$key: seed ID, $value: number of reads in the OTU
	my %derep_info = %{$_[2]}; #$key: sequence ID chosen after dereplication, $value: corresponding unique reads
	my $cluster = "";
	my $ID = "";
	my $new_cluster = "";
	my @tmp = "";
	my $seed = "";

	LOOP01:foreach $cluster (keys(%corre))
	{
		if(keys(%derep_info) == 0) { next LOOP01; }
		#We catch the seed ID as needed in a further step
		$seed = $corre{$cluster};
		#We store in @tmp the list of each OTU
		@tmp = split(" ", $cluster);
		#Needed empty variable
		$new_cluster = "";
		#Foreach ID sequence in the cluster
		LOOP02:foreach $ID (@tmp)
		{
			#If this one exists and can be replaced
			if(exists $derep_info{$ID})
			{
				#We replace the sequence by its corresponding info and delete the data in %delete_info
				$new_cluster .= $derep_info{$ID}." ";
				delete($derep_info{$ID});
				next LOOP02;
			}
			#Or, we let the $ID as it is in the $new_cluster
			else { $new_cluster .= $ID." "; }
		}
		#We finally modify the data in %corre
		$corre{$new_cluster} .= $seed;
		delete($corre{$cluster});
	}
	#We finally keep only the three needed data (one array, two hashes).
	return(\%corre, \%clust);
}
#End of function
###############################################################################

###############################################################################
#FUNCTION:	DefineSamplingDistribution
#USAGE:			DefineSamplingDistribution(\%corre)
#ARGUMENTS:	-$_[0]: reference to %corre hash (seed ID linked to clustering)
#DETAILS:	This function was developed to treat the information from
#		each cluster using the %corre as input, and give back a line
#		of description (nbreads_sampleID|nbreads_sample_ID...) to
#		describe the composition of each OTU in the database.
#RELEASED:	16/10/2017
###############################################################################
sub DefineSamplingDistribution
{
	#Defined variables
	my %corre = %{$_[0]}; #$key: line of clustering composition, $value: ID seed
	my %glob_distrib = (); #$key: seed ID for the cluster, $value : line of info for db
	my %clust_distrib = (); #$key: sample ID, $value: nb reads for this sample
	my $cluster = "";
	my $ID = "";
	my @tmp = ();

	#For each line of clustering found in %corre given as input
	foreach $cluster (keys(%corre))
	{
		#We split the line and treat each unique read
		@tmp = split(" ", $cluster);
		foreach $ID (@tmp)
		{
			#We catch the origin of the read and its dereplication info
			$ID =~ m/(\d+)\|.*\_(\d+)\_.*/;
			#We store the info in %clust_distrib
			$clust_distrib{$1} += $2;
		}
		#We then catch all sample IDs, sort and store them in @tmp
		@tmp = keys(%clust_distrib);
		@tmp = sort(@tmp);
		#And then, for each sample ID
		foreach $ID (@tmp)
		{
			#We then create the descriptive line for the database based on the info from %clust_distrib
			$glob_distrib{$corre{$cluster}} .= $clust_distrib{$ID}."_".$ID."|";
		}
		#We finally clean the not needed characters and start the analysis of another cluster
		$glob_distrib{$corre{$cluster}} =~ s/\|$//;
		%clust_distrib = ();
	}
	#We then return the details for eahc cluster to the main program
	return(\%glob_distrib);
}
#End of function
###############################################################################

###############################################################################
#FUNCTION:	ExtractAndStoreSeedSeqs
#USAGE:	ExtractAndStoreSeedSeq(ARGUMENTS)
#ARGUMENTS:	-$_[0]: sequence filename given by the user and checked
#						-$_[1]: reference to %IDs hash (seed ID linked to OTU ID)
#						-$_[2]: name of the created database folder
#						-$_[3]: reference to @choices array (user choices)
#DETAILS:	This function is dedicated to the search and storage of the seed
#		sequences to create the complete database. Based on several inputs
#		(sequence filename, %IDs hash, folder name of the database and user
#		choices (@choices), this function seek all needed reads,
#		check if nothing is lost, and create the dbreads file.
#RELEASED:	16/10/2017
###############################################################################
sub ExtractAndStoreSeedSeqs
{
	#Defined variables
	my $seq_file = $_[0];
	my %IDs = %{$_[1]};
	my $db_folder = $_[2];
	my @choices = @{$_[3]};
	my @seeds = ();
	my %sequences = (); # $key: seed ID, $value: sequence itself
	my $seed = "";
	my $db_file = "dbreads_".$choices[0]."_".$choices[4]."_C".$choices[2]."_T".$choices[3];

	#First, we used all given seeds to define our keys' hash
	@seeds = keys(%IDs);
	foreach $seed (@seeds) { $sequences{$seed} = ""; }
	#Then, we open and read the file to find all corresponding values
	open(F_SEQ, "$seq_file") || die "==> Can't open the $seq_file given as input\n";
	#For each line of the treated file
	while(<F_SEQ>)
	{
		#If it's a descriptive line, we check if the ID is a seed, if yes, we also
		#store the reads in %sequences.
		if($_ =~ m/^>\d*\|(.*)\n/)
		{
			if(exists($sequences{$1})) { $sequences{$1} = <F_SEQ>; }
		}
	}
	close(F_SEQ);
	#At the end, we check if we found all seed sequences needed
	foreach $seed (keys(%sequences))
	{
		if($sequences{$seed} eq "") { die "Can't find the $seed sequence in the $seq_file given as input.\n"; }
	}

	#Based on user choices we define the wanted filename
	if($choices[1] eq "YES") { $db_file .= "_G"; }
	else { $db_file .= "_F"; }
	open(DB_READS, ">$db_folder/$db_file") || die "==> Can't create the database dbreads file in the $db_folder folder.\n";
	#Then, using all seeds, we fill the dbreads file dedicated to seed OTUs
	foreach $seed (keys(%sequences))
	{
		$IDs{$seed} = "OTU".sprintf("%09d", $IDs{$seed});
		printf(DB_READS ">$IDs{$seed}\n$sequences{$seed}");
	}
	close(DB_READS);
	#As we modified the seed IDs, we returned it to the main program with the db_file created
	return(\%IDs, $db_file);
}
#End of function
################################################################################

###############################################################################
#FUNCTION:	ExtractTaxonomicData
#USAGE:			ExtractTaxonomicData()
#ARGUMENTS:	-$_[0]: taxonomy filename given by the user and checked
#
#DETAILS:		This function was developed to find, format and extract all
#		taxonomic details for each OTU. This function needs only the
#		taxonomic filename as input, and gives back as output a hash table
#		with OTU ID as key and as value the details of taxonomy.
#RELEASED:	16/10/2017
###############################################################################
sub ExtractTaxonomicData
{
	#Defined variables
	my $taxo_file = $_[0];
	my @tmp = ();
	my @tmp2 = ();
	my %Taxo_ID = (); #$key: OTU ID, $value: taxo details
	my $ID = "";
	my $element = "";
	my $max = 0;
	my $kept_taxo = "";
	my $taxo = "";
	my $sum = 0;
	my $i = 0;

	#We open the taxonomic file
	open(F_TAXO, "$taxo_file") || die "==> Can't open the $taxo_file given as input\n";
	#For each line of the treated file
	LOOP01:while(<F_TAXO>)
	{
		#First we split the line based on tabulations
		@tmp = split("\t", $_);
		#Check if the we have the good start of line (OTU and/or the OTU ID)
		if(($tmp[0] =~ m/(^\d+)/)||($tmp[0] =~ m/^OTU(\d+)/))
		{
			#We keep the details of the OTU id for the main program
			$ID = $1;
			#Then, for each element, starting with the phylum level
			LOOP02:for($i=1;$i<=$#tmp;$i++)
			{
				#We check the number of taxonomic names
				@tmp2 = split(" ", $tmp[$i]);
				#If we have one name, we catch the given name and store it
				if($#tmp2 == 0)
				{
					#Here, we search for the name, and avoid the analysis of the number of reads,
					#as we have only one taxonomic name
					if(($tmp2[0] =~ m/(.*)\(\d+\)/)||($tmp2[0] =~ m/(.*)/))
					{
						$taxo .= "1.00_".$1."|";
						next LOOP02;
					}
				}
				#Needed empty variables
				$sum = 0;
				$max = 0;
				$kept_taxo = "";
				#If we have several elements for the taxo, we chose the best one, and compute the
				#percent of values
				LOOP03:foreach $element (@tmp2)
				{
					#We catch the taxonomic name and the number of affiliated sequences
					$element =~ m/(.*)\((\d+)\)/;
					#We store only the highest level
					if($max <= $2)
					{
						#Particular cases with 50% Unknown, 50 Others
						if(($max == $2)&&($1 eq "Unknown")) { $sum += $2; next LOOP03; }
						if(($max == $2)&&($kept_taxo eq "Unknown"))
						{
							$kept_taxo = $1;
							$max = $2;
						}
						$kept_taxo = $1;
						$max = $2;
					}
					#We count the number of reads in this OTU, to define the needed percentage
					$sum += $2;
				}
				#We compute the percentage for this OTU with this taxonomic name
				$max = sprintf("%0.2f", $max/$sum);
				#We define the details in $taxo
				$taxo .= $max."_".$kept_taxo."|";
			}
		}
		#We finally store the data in the dedicated hash table
		$taxo =~ s/\|$//;
		$Taxo_ID{"OTU".sprintf("%09d", $ID)} = $taxo;
		$taxo = "";
	}
	close(F_TAXO);
	return(\%Taxo_ID);
}
#End of function
################################################################################

###############################################################################
#FUNCTION:	DefineDatabaseFile
#USAGE:			DefineDatabaseFile()
#ARGUMENTS:	-$_[0]: ref to %clust (seed ID linked to number of reads in the OTU),
#	-$_[1]: ref to %glob_distrib (seed ID for the cluster linked to line of info for db),
# -$_[2]: ref to %IDs (seed ID linked to cluster ID),
# -$_[3]: ref to %Taxo_IDs (OTU ID linked to taxonomic details),
# -$_[4]: the folder defined to store the database,
# -$_[5]: ref to the array containing the user choices.
#
#DETAILS:	This function was developed to define the file containing all data
#needed the database creation (file containing all details for each seed sequence)
#RELEASED:	16/10/2017
###############################################################################
sub DefineDatabaseFile
{
	#Defined variables
	my %clust = %{$_[0]};
	my %glob_distrib = %{$_[1]};
	my %IDs = %{$_[2]};
	my %Taxo_IDs = %{$_[3]};
	my $db_folder = $_[4];
	my @choices = @{$_[5]};
	my %Final_data = ();
	my @tab = ();
	my $key = "";
	my $db_file = "database_".$choices[0]."_".$choices[4]."_C".$choices[2]."_T".$choices[3];

	#Based on user choices we define the wanted filename
	if($choices[1] eq "YES") { $db_file .= "_G"; }
	else { $db_file .= "_F"; }
	#We reorganize using the different hashes the data stored in the database
	foreach $key (keys(%clust)) { $Final_data{$IDs{$key}} = "$clust{$key}\t$Taxo_IDs{$IDs{$key}}\t$glob_distrib{$key}\n"; }
	#Then, based on the keys (reorganized by alphabetiacal order)
	@tab = keys(%Final_data);
	@tab = sort { length($b) <=> length($a) || $a cmp $b } @tab;
	#We store in the database file the needed data to define completely the database itself
	open(F_OUT, ">$db_folder/$db_file") || die "==> Can't create the database file in the $db_folder folder.\n";
	foreach $key (@tab) { printf(F_OUT "$key\t$Final_data{$key}"); }
	close(F_OUT);
}
#End of function
################################################################################
1;
